#!/use/bin/env python
# coding=utf-8
# @Author  : Shuhao Liu
# @Time    : 2025/6/16 21:01 
# @File    : collect_auxiliary.py
import pathlib
import os
import ssl

import wget

import sagea
from sagea.pysrc.auxiliary.FileTool import FileTool
from sagea.pysrc.data_collection.processing_bar import bar


def collect_auxiliary(local_dir):
    """download and unzip auxiliary and validation data from https://zenodo.org/records/15675307."""

    """prepare"""
    local_dir = pathlib.Path(local_dir)

    """download data"""
    url = f"https://zenodo.org/records/15675307/files/sagea_data.zip?download=1"

    local_path = local_dir / "sagea_data.zip"

    print(f"downloading: {url}")

    wget.download(url, str(local_path), bar=bar)

    print()

    print(f"unzipping: {local_path} ...")
    """unzip"""

    FileTool.un_zip(local_path, local_dir)

    os.remove(local_path)
    print("done!\n")


def collect_ddk_data(local_dir=None):
    """download and unzip auxiliary and validation data from https://zenodo.org/records/15679042."""

    if local_dir is None:
        local_dir = sagea.Preference.Config.aux_data_dir

    """prepare"""
    local_dir = pathlib.Path(local_dir)

    """download data"""
    url = f"https://zenodo.org/records/15679042/files/ddk_data.zip?download=1"

    local_path = local_dir / "sagea_data.zip"

    print(f"downloading DDK data from: {url}")

    wget.download(url, str(local_path), bar=bar)

    print()

    print(f"unzipping: {local_path} ...")
    """unzip"""

    FileTool.un_zip(local_path, local_dir)

    os.remove(local_path)
    print("done!\n")


if __name__ == "__main__":
    pass



#!/use/bin/env python
# coding=utf-8
# @Author  : Shuhao Liu
# @Time    : 2025/6/16 21:03 
# @File    : processing_bar.py

def bar(current, total, width):
    progress = current / total
    wid = 30
    bar_style = '#' * (int(wid * progress) + 1) + '-' * (wid - int(wid * progress) - 1)

    print(f'\r|{bar_style}| {"%.2f" % (current / total * 100)}%', end='')



import numpy as np


class ReplaceLowDegreeConfig:
    def __init__(self):
        self.replace_c00 = False

        self.replace_c10 = False
        self.replace_c11 = False
        self.replace_s11 = False

        self.replace_c20 = False
        self.replace_c30 = False

    def set_replace_c00(self, replace_or_not: bool = True):
        self.replace_c00 = replace_or_not

        return self

    # def set_replace_deg1(self, replace_or_not: bool = True):
    #     self.replace_c10 = replace_or_not
    #     self.replace_c11 = replace_or_not
    #     self.replace_s11 = replace_or_not
    #
    #     return self

    def set_replace_c10(self, replace_or_not: bool = True):
        self.replace_c10 = replace_or_not

        return self

    def set_replace_c11(self, replace_or_not: bool = True):
        self.replace_c11 = replace_or_not

        return self

    def set_replace_s11(self, replace_or_not: bool = True):
        self.replace_s11 = replace_or_not

        return self

    def set_replace_c20(self, replace_or_not: bool = True):
        self.replace_c20 = replace_or_not

        return self

    def set_replace_c30(self, replace_or_not: bool = True):
        self.replace_c30 = replace_or_not

        return self


class ReplaceLowDegree:
    def __init__(self):
        self.configuration = ReplaceLowDegreeConfig()

        self.low_degrees = None

    def set_low_degrees(self, low_degrees: dict):
        """
        :return:
        """
        self.low_degrees = low_degrees

        return self

    def _get_ymd_array(self, dates):
        ymd_list = []
        for i in range(len(dates)):
            year = str(dates[i].year)
            month = str(dates[i].month).rjust(2, '0')
            day = str(dates[i].day).rjust(2, '0')

            ymd_str = year + month + day
            ymd_list.append(int(ymd_str))

        return np.array(ymd_list)

    def apply_to(self, cqlm, sqlm, begin_dates, end_dates, dev=False):
        """
        :param cqlm:
        :param sqlm:
        :param begin_dates: list of datetime.date
        :param end_dates: list of datetime.date
        :param dev: replace deviation or not
        :return:
        """
        begin_times = self._get_ymd_array(begin_dates)
        end_times = self._get_ymd_array(end_dates)

        begin_times = np.array([int(begin_times[i]) for i in range(len(begin_times))])
        end_times = np.array([int(end_times[i]) for i in range(len(end_times))])

        if self.configuration.replace_c00:
            assert 'c00' in self.low_degrees.keys()

            replace_times = self._get_ymd_array(self.low_degrees['c00'][0])[:, None]
            replace_values = self.low_degrees['c00'][1]

            where = np.where(((begin_times <= replace_times) * (replace_times <= end_times)))

            cqlm[where[1], 0, 0] = replace_values[where[0]]

        if self.configuration.replace_c10:
            # assert 'c10' in self.low_degrees.keys()
            if dev:
                assert 'c10_dev' in self.low_degrees.keys()

                replace_times = self._get_ymd_array(self.low_degrees['c10_dev'][0])[:, None]
                replace_values = self.low_degrees['c10_dev'][1]

            else:
                assert 'c10' in self.low_degrees.keys()

                replace_times = self._get_ymd_array(self.low_degrees['c10'][0])[:, None]
                replace_values = self.low_degrees['c10'][1]

            where = np.where(((begin_times <= replace_times) * (replace_times <= end_times)))

            cqlm[where[1], 1, 0] = replace_values[where[0]]

        if self.configuration.replace_c11:
            # assert 'c11' in self.low_degrees.keys()
            if dev:
                assert 'c11_dev' in self.low_degrees.keys()

                replace_times = self._get_ymd_array(self.low_degrees['c11_dev'][0])[:, None]
                replace_values = self.low_degrees['c11_dev'][1]

            else:
                assert 'c11' in self.low_degrees.keys()

                replace_times = self._get_ymd_array(self.low_degrees['c11'][0])[:, None]
                replace_values = self.low_degrees['c11'][1]

            where = np.where(((begin_times <= replace_times) * (replace_times <= end_times)))

            cqlm[where[1], 1, 1] = replace_values[where[0]]

        if self.configuration.replace_s11:
            # assert 's11' in self.low_degrees.keys()
            if dev:
                assert 's11_dev' in self.low_degrees.keys()

                replace_times = self._get_ymd_array(self.low_degrees['s11_dev'][0])[:, None]
                replace_values = self.low_degrees['s11_dev'][1]

            else:
                assert 's11' in self.low_degrees.keys()

                replace_times = self._get_ymd_array(self.low_degrees['s11'][0])[:, None]
                replace_values = self.low_degrees['s11'][1]

            where = np.where(((begin_times <= replace_times) * (replace_times <= end_times)))

            sqlm[where[1], 1, 1] = replace_values[where[0]]

        if self.configuration.replace_c20:
            if dev:
                assert 'c20_dev' in self.low_degrees.keys()

                replace_times = self._get_ymd_array(self.low_degrees['c20_dev'][0])[:, None]
                replace_values = self.low_degrees['c20_dev'][1]

            else:
                assert 'c20' in self.low_degrees.keys()

                replace_times = self._get_ymd_array(self.low_degrees['c20'][0])[:, None]
                replace_values = self.low_degrees['c20'][1]

            where = np.where(((begin_times <= replace_times) * (replace_times <= end_times)))

            cqlm[where[1], 2, 0] = replace_values[where[0]]

        if self.configuration.replace_c30:
            # assert 'c30' in self.low_degrees.keys()
            if dev:
                assert 'c30_dev' in self.low_degrees.keys()

                replace_times = self._get_ymd_array(self.low_degrees['c30_dev'][0])[:, None]
                replace_values = self.low_degrees['c30_dev'][1]

            else:
                assert 'c30' in self.low_degrees.keys()

                replace_times = self._get_ymd_array(self.low_degrees['c30'][0])[:, None]
                replace_values = self.low_degrees['c30'][1]

            where = np.where(((begin_times <= replace_times) * (replace_times <= end_times)))

            cqlm[where[1], 3, 0] = replace_values[where[0]]

        return cqlm, sqlm



"""
@Company: CGE-HUST, Wuhan, China
@Author: Yang Fan
@Contact: yfan_cge@hust.edu.cn
@Modify Time:2020/10/28 下午4:56
@Description:
"""

import os
import warnings

import numpy as np


class LoadSH:
    """
    This class is a base class that deals with the gravity models reading.
    """

    def __init__(self):
        self.product_type = None
        self.modelname = None
        self.GM = None
        self.Radius = None
        self.maxDegree = None
        self.zero_tide = None
        self.errors = None
        self.norm = None
        self.date_begin = None
        self.date_end = None
        self.date_middle = None
        self.unusedDays = None

        self._sigmaC, self._sigmaS = None, None
        self._C, self._S = None, None
        pass

    def load(self, fileIn: str):
        return self

    def getCS(self, Nmax):

        assert Nmax >= 0

        end = int((Nmax + 1) * (Nmax + 2) / 2)

        if end > len(self._C):
            warnings.warn('Nmax is too big')

        return self._C[0:end].copy(), self._S[0:end].copy()

    def getSigmaCS(self, Nmax):

        assert Nmax >= 0

        end = int((Nmax + 1) * (Nmax + 2) / 2)

        if end > len(self._C):
            warnings.warn('Nmax is too big')

        end = int((Nmax + 1) * (Nmax + 2) / 2)

        return self._sigmaC[0:end].copy(), self._sigmaS[0:end].copy()

    def replace(self):
        """
        :return:
        """
        return self._C, self._S, self._sigmaC, self._sigmaS


class SimpleSH(LoadSH):
    """
    specified to the ones in the simplest format (No header) like below:

    0  0  c s
    1  0  c s
    1  1  c s
    """

    def __init__(self):
        LoadSH.__init__(self)

    def load(self, fileIn: str):

        with open(fileIn) as f:
            content = f.readlines()
            pass

        l, m, C, S = [], [], [], []

        for item in content:
            value = item.split()
            if len(value) == 0: continue

            l.append(value[0])
            m.append(value[1])
            C.append(value[2])
            S.append(value[3])

        l = np.array(l).astype(int)
        m = np.array(m).astype(int)
        C = np.array(C).astype(float)
        S = np.array(S).astype(float)

        n = np.round(np.sqrt(len(l) * 2)) - 1
        assert len(l) == int((n + 1) * (n + 2) / 2)

        self._C = np.zeros(len(l))
        self._S = np.zeros(len(l))
        self._sigmaC = np.zeros(len(l))
        self._sigmaS = np.zeros(len(l))

        self._C[(l * (l + 1) / 2 + m).astype(np.int)] = C
        self._S[(l * (l + 1) / 2 + m).astype(np.int)] = S

        return self


class Gif48(LoadSH):
    """
    specified to read gif48 gravity fields. Plus, the file that formats the same as GIF48 can be read as well.
    """

    def __init__(self):
        LoadSH.__init__(self)
        # self.__sigmaC, self.__sigmaS = None, None
        # self.__C, self.__S = None, None

    def load(self, fileIn: str):
        """
        load gif48 fields
        :param fileIn: gif48 file and its path
        :return:
        """

        flag = 0

        with open(fileIn) as f:
            content = f.readlines()
            pass

        flag = self.__header(content)

        self.__read(content[flag + 1:])

        return self

    def __header(self, content: list):
        """
        Read and identify the header of the gif48 gravity fields file
        :param content: content of the file
        :return: the line of "end_of_header"
        """

        flag = 0

        for i in range(len(content)):
            value = content[i].split()
            if len(value) == 0: continue

            if value[0] == 'product_type':
                self.product_type = value[1]
            elif value[0] == 'modelname':
                self.modelname = value[1]
            elif value[0] == 'earth_gravity_constant':
                self.GM = float(value[1])
            elif value[0] == 'radius':
                self.Radius = float(value[1])
            elif value[0] == 'max_degree':
                self.maxDegree = int(value[1])
            elif value[0] == 'errors':
                self.errors = value[1]
            elif value[0] == 'norm':
                self.norm = value[1]
            elif value[0] == 'tide_system':
                self.zero_tide = (value[1] == 'zero_tide')
            elif value[0] == 'end_of_head':
                flag = i
                break

        return flag

    def __read(self, content: list):
        """
        Start reading gif48 like line by line like below
        key    L    M    C    S    sigma C    sigma S
        :param content:
        :return: all the C, S coefficients in order
        """
        l, m, C, S, sigmaC, sigmaS = [], [], [], [], [], []

        for item in content:
            value = item.split()
            if len(value) == 0: continue

            l.append(value[1])
            m.append(value[2])
            C.append(value[3])
            S.append(value[4])
            sigmaC.append(value[5])
            sigmaS.append(value[6])

        l = np.array(l).astype(int)
        m = np.array(m).astype(int)
        C = np.array(C).astype(float)
        S = np.array(S).astype(float)
        sigmaC = np.array(sigmaC).astype(float)
        sigmaS = np.array(sigmaS).astype(float)

        if self.maxDegree is not None:
            assert len(l) == int((self.maxDegree + 1) * (self.maxDegree + 2) / 2)
        else:
            n = np.round(np.sqrt(len(l) * 2)) - 1
            assert len(l) == int((n + 1) * (n + 2) / 2)

        self._C = np.zeros(len(l))
        self._S = np.zeros(len(l))
        self._sigmaC = np.zeros(len(l))
        self._sigmaS = np.zeros(len(l))

        self._C[(l * (l + 1) / 2 + m).astype(int)] = C
        self._S[(l * (l + 1) / 2 + m).astype(int)] = S
        self._sigmaC[(l * (l + 1) / 2 + m).astype(int)] = sigmaC
        self._sigmaS[(l * (l + 1) / 2 + m).astype(int)] = sigmaS


class Level2Gravity(LoadSH):
    """
    This is used to read level 2 temporal gravity fields from official data centers: CSR, JPL and GFZ
    """

    def load(self, fileIn: str):
        """
        Read one file using given directory and filename. This function asks for an exact filename.
        :param fileIn:
        :return:
        """

        flag = 0

        with open(fileIn) as f:
            content = f.readlines()
            pass

        flag = self.__header(content)

        self.__read(content[flag + 1:])

        return self

    def __header(self, content: list):
        """
        Read and identify the header of the gif48 gravity fields file
        :param content: content of the file
        :return: the line of "end_of_header"
        """

        flag = 0

        for i in range(len(content)):
            value = content[i].split(':')

            if len(value) == 0: continue
            if 'End of YAML header' in value[0]:
                flag = i
                break

            if 'time_coverage_start' in value[0]:
                self.date_begin = value[1].strip().split('T')[0]
            if 'time_coverage_end' in value[0]:
                self.date_end = value[1].strip().split('T')[0]
            if 'unused_days' in value[0]:
                self.unusedDays = value[1].strip().strip()[1:-1].split(',')

            if 'title' in value[0]:
                self.product_type = value[1].strip()
            elif 'product_version' in value[0]:
                self.modelname = value[1].strip()
            elif 'mean_equator_radius' in value[0]:
                self.Radius = float(content[i + 3].split(':')[1])
            elif 'earth_gravity_param' in value[0]:
                self.GM = float(content[i + 3].split(':')[1])
            elif 'normalization' in value[0]:
                self.norm = value[1].strip()
            elif 'permanent_tide_flag' in value[0]:
                self.zero_tide = value[1].strip()
            elif 'dimensions' in value[0]:
                self.maxDegree = int(content[i + 1].split(':')[1])
            elif 'comments' in value[0]:
                self.errors = value[1].split()[0]

        '''compute the middle date in terms of month: for the later time-series analysis'''
        '''Time is not evenly distributed'''
        be = int(self.date_begin[:4]) + ((int(self.date_begin[5:7])-1)*(365.25/12)+ int(self.date_begin[8:10]))/365.25
        en= int(self.date_end[:4]) + (
                    (int(self.date_end[5:7]) - 1) * (365.25 / 12) + int(self.date_end[8:10])) / 365.25
        self.date_middle = (be+en)/2*12

        return flag

    def __read(self, content: list):
        """
        Start reading gif48 like line by line like below
        key    L    M    C    S    sigma C    sigma S
        :param content:
        :return: all the C, S coefficients in order
        """
        l, m, C, S, sigmaC, sigmaS = [], [], [], [], [], []

        for item in content:
            value = item.split()
            if len(value) == 0: continue

            l.append(value[1])
            m.append(value[2])
            C.append(value[3])
            S.append(value[4])
            sigmaC.append(value[5])
            sigmaS.append(value[6])

        l = np.array(l).astype(int)
        m = np.array(m).astype(int)
        C = np.array(C).astype(float)
        S = np.array(S).astype(float)
        sigmaC = np.array(sigmaC).astype(float)
        sigmaS = np.array(sigmaS).astype(float)

        if self.maxDegree is not None:
            assert len(l) == int((self.maxDegree + 1) * (self.maxDegree + 2) / 2)
        else:
            n = np.round(np.sqrt(len(l) * 2)) - 1
            assert len(l) == int((n + 1) * (n + 2) / 2)

        self._C = np.zeros(len(l))
        self._S = np.zeros(len(l))
        self._sigmaC = np.zeros(len(l))
        self._sigmaS = np.zeros(len(l))

        self._C[(l * (l + 1) / 2 + m).astype(int)] = C
        self._S[(l * (l + 1) / 2 + m).astype(int)] = S
        self._sigmaC[(l * (l + 1) / 2 + m).astype(int)] = sigmaC
        self._sigmaS[(l * (l + 1) / 2 + m).astype(int)] = sigmaS


def LoadGsmByYear(localDir: str, begin: str, end: str, opt='60'):
    """
    load files by years
    :param opt: degree/order = 60 or 90
    :param localDir: directory where temporal gravity fields are saved
    :param begin: e.g., '2009'
    :param end: e.g., '2010'
    :return: a list of time-series GSM sorted by date
    """
    res = []
    assert opt in ['60', '90']

    id = None
    if opt == '60':
        id = 'BA01'
    elif opt == '90':
        id = 'BB01'

    target = None
    years = np.arange(int(begin), int(end) + 1)

    for root, dirs, files in os.walk(localDir):
        for name in files:
            str_date = name.split('_')[1].split('-')
            start = str_date[0]

            if ('GSM' in name) and (id in name) and (int(start[:4]) in years):
                target = os.path.join(root, name)
                res.append(Level2Gravity().load(target))
                pass

    '''
    sort the list by date
    '''

    dates = [x.date_begin for x in res]
    index = sorted(range(len(dates)), key=lambda k: dates[k])
    ress = [res[i] for i in index]

    return ress


def demo1():
    l2sh = Level2Gravity().load('../data/L2_SH_products/CSR/GSM/BA01/2002/GSM-2_2002095-2002120_GRAC_UTCSR_BA01_0600')
    pass


def demo2():
    ts = LoadGsmByYear(localDir='../data/L2_SH_products/CSR', begin='2002', end='2020', opt='90')
    for el in ts:
        print(el.date_begin)
    pass


if __name__ == '__main__':
    demo2()



import copy

import numpy as np

from pysrc.data_class.SHC import SHC
from pysrc.auxiliary.aux_tool.TimeTool import TimeTool


class GIACorrectionSpectralConfig:
    def __init__(self):
        self.__gia_trend = None
        self.__times = None

    def set_gia_trend(self, gia: SHC):
        self.__gia_trend = gia

        return self

    def get_gia_trend(self):
        return self.__gia_trend

    def set_times(self, times):
        """
        :param times: list of datetime.date
        """
        self.__times = times
        return self

    def get_times(self):
        return self.__times


class GIACorrectionSpectral:
    def __init__(self):
        self.configuration = GIACorrectionSpectralConfig()

    def __get_year_fractions(self):
        times_list = self.configuration.get_times()  # list datetime.date

        year_frac = TimeTool.convert_date_format(times_list,
                                                 input_type=TimeTool.DateFormat.ClassDate,
                                                 output_type=TimeTool.DateFormat.YearFraction)

        year_frac = np.array(year_frac)

        return year_frac

    def apply_to(self, shc):
        cs_gia_trend = self.configuration.get_gia_trend().value[0]
        year_frac = self.__get_year_fractions()
        year_frac -= np.mean(year_frac)

        cs_each_times = year_frac[:, None] @ cs_gia_trend[None, :]

        shc_new = copy.deepcopy(shc)
        shc_new.value -= cs_each_times

        return shc_new



import numpy as np

from sagea.pysrc.auxiliary.Preference import Dimension, LoveNumberType
from sagea.pysrc.auxiliary.Preference import GeoConstants
from sagea.pysrc.auxiliary.LoveNumber import LoveNumber


class ConvertSHCConfig:
    def __init__(self):
        self.ln = None

        self.input_field_type = Dimension.Geopotential
        self.output_field_type = Dimension.EWH

    def set_input_type(self, field_type: Dimension):
        self.input_field_type = field_type
        return self

    def set_output_type(self, field_type: Dimension):
        self.output_field_type = field_type
        return self

    def set_Love_number(self, ln):
        self.ln = ln
        return self


class ConvertSHC:
    def __init__(self):
        self.configuration = ConvertSHCConfig()

    def config(self, config: ConvertSHCConfig):
        self.configuration = config
        return self

    def apply_to(self, cs1d):

        length_of_cs1d = np.shape(cs1d)[-1]
        lmax = int(np.sqrt(length_of_cs1d) - 1)

        convert_array = self._get_convert_array_to_dimensionless(
            self.configuration.input_field_type,
            lmax
        ) * self._get_convert_array_from_dimensionless_to(
            self.configuration.output_field_type, lmax
        )
        # [k1, k2, ...,]

        convert_weight_cs1d = np.array([])
        for i in range(lmax + 1):
            convert_weight_cs1d = np.concatenate([convert_weight_cs1d, [convert_array[i]] * (2 * i + 1)])

        cs1d_converted = cs1d * convert_weight_cs1d

        return cs1d_converted

    def _get_convert_array_to_dimensionless(self, field_type: Dimension, lmax):
        """
        return: [k1, k2, ..., kl, ...]
        """
        assert self.configuration.ln is not None
        ln = self.configuration.ln

        convert_mat = np.ones((lmax + 1,))

        density_water = GeoConstants.density_water
        density_earth = GeoConstants.density_earth
        radius_e = GeoConstants.radius_earth
        GM = GeoConstants.GM
        g_wmo = GeoConstants.g_wmo

        if field_type is Dimension.Geopotential:
            pass

        elif field_type is Dimension.EWH:
            ln = ln[:lmax + 1]
            kl = np.array([(1 + ln[n]) / (2 * n + 1) for n in range(len(ln))]) * 3 * density_water / (
                    radius_e * density_earth)

            convert_mat *= kl

        elif field_type is Dimension.Density:
            ln = ln[:lmax + 1]
            kl = np.array([(1 + ln[n]) / (2 * n + 1) for n in range(len(ln))]) * 3 / (radius_e * density_earth)

            convert_mat *= kl

        elif field_type is Dimension.Geoid:
            convert_mat /= radius_e

        elif field_type is Dimension.Gravity:
            kl = np.array([n - 1 for n in range(lmax + 1)]) * (GM / radius_e ** 2)
            convert_mat /= kl

        elif field_type is Dimension.VerticalDisplacement:
            lnh, lnl = self.__get_love_number_h_and_l(lmax)

            ln = ln[:lmax + 1]
            kl = np.array([lnh[n] / (1 + ln[n]) for n in range(len(ln))]) * radius_e

            convert_mat /= kl

        elif field_type is Dimension.Pressure:
            termI = np.arange(lmax + 1)
            term = 2 * termI + 1.
            ln = ln[:lmax + 1]

            kl = (radius_e * density_earth / 3) * (term / (1 + ln)) * g_wmo

            convert_mat /= kl

        else:
            raise Exception

        return convert_mat

    def _get_convert_array_from_dimensionless_to(self, field_type: Dimension, lmax):
        """
        return: [k1, k2, ..., kl, ...]
        """

        assert self.configuration.ln is not None
        ln = self.configuration.ln

        density_water = GeoConstants.density_water
        density_earth = GeoConstants.density_earth
        radius_e = GeoConstants.radius_earth
        GM = GeoConstants.GM
        g_wmo = GeoConstants.g_wmo

        convert_mat = np.ones((lmax + 1,))

        if field_type is Dimension.Geopotential:
            pass

        elif field_type is Dimension.EWH:
            ln = ln[:lmax + 1]
            kl = np.array([(2 * n + 1) / (1 + ln[n]) for n in range(len(ln))]) * radius_e * density_earth / (
                    3 * density_water)

            convert_mat *= kl

        elif field_type is Dimension.Density:
            ln = ln[:lmax + 1]
            kl = np.array([(2 * n + 1) / (1 + ln[n]) for n in range(len(ln))]) * radius_e * density_earth / 3

            convert_mat *= kl

        elif field_type is Dimension.Geoid:
            convert_mat *= radius_e

        elif field_type is Dimension.Gravity:
            kl = np.array([n - 1 for n in range(lmax + 1)]) * (GM / radius_e ** 2)
            convert_mat *= kl

        elif field_type is Dimension.VerticalDisplacement:
            lnh, lnl = self.__get_love_number_h_and_l(lmax)

            ln = ln[:lmax + 1]
            kl = np.array([lnh[n] / (1 + ln[n]) for n in range(len(ln))]) * radius_e

            convert_mat *= kl

        elif field_type is Dimension.Pressure:
            termI = np.arange(lmax + 1)
            term = 2 * termI + 1.
            ln = ln[:lmax + 1]

            kl = (radius_e * density_earth / 3) * (term / (1 + ln)) * g_wmo

            convert_mat *= kl

        elif field_type in (
                Dimension.HorizontalDisplacementNorth, Dimension.HorizontalDisplacementEast):
            lnh, lnl = self.__get_love_number_h_and_l(lmax)

            ln = ln[:lmax + 1]
            kl = np.array([lnl[n] / (1 + ln[n]) for n in range(len(ln))]) * radius_e

            convert_mat *= kl

        else:
            raise Exception

        return convert_mat

    @staticmethod
    def __get_love_number_h_and_l(lmax):
        LN = LoveNumber()

        LN.configuration.set_lmax(lmax).set_type(LoveNumberType.VerticalDisplacement)  # h
        ln_h = LN.get_Love_number()

        LN.configuration.set_lmax(lmax).set_type(LoveNumberType.HorizontalDisplacement)  # l
        ln_l = LN.get_Love_number()

        return ln_h, ln_l



import numpy as np
# from scipy import optimize

from sagea.pysrc.auxiliary.TimeTool import TimeTool
from sagea.pysrc.auxiliary.MathTool import MathTool


class SeismicCorrectionConfig:
    def __init__(self):
        self.__earthquake_list = None
        # dict, {'name': {'lat_range': [., .], 'lon_range': [., .], 'teq': [. ,], 'tau': [. ,], }}

        self.__times = None

    def set_earthquakes(self, events: dict):
        """
        set earthquake events from dict or .json file
        """
        assert issubclass(type(events), dict)

        self.__set_earthquakes_from_dict(events)

        return self

    def __set_earthquakes_from_dict(self, events: dict):
        self.__earthquake_list = events

        return self

    def get_earthquakes(self):
        return self.__earthquake_list

    def set_times(self, times):
        """
        :param times: list of datetime.date
        """
        self.__times = times
        return self

    def get_times(self):
        return self.__times


class SeismicCorrection:
    def __init__(self):
        self.configuration = SeismicCorrectionConfig()

    def __get_year_fractions(self):
        times_list = self.configuration.get_times()  # list datetime.date

        year_frac = TimeTool.convert_date_format(times_list,
                                                 input_type=TimeTool.DateFormat.ClassDate,
                                                 output_type=TimeTool.DateFormat.YearFraction)

        year_frac = np.array(year_frac)

        return year_frac

    def apply_to(self, gqij, lat, lon):
        earthquakes = self.configuration.get_earthquakes()
        for key in earthquakes:
            '''get lat and lon index of grid.data'''
            assert len(earthquakes[key]['lat_range']) == 2
            assert len(earthquakes[key]['lon_range']) == 2
            assert len(earthquakes[key]['teq']) in (1, 2)
            assert len(earthquakes[key]['teq']) == len(earthquakes[key]['tau'])

            lat_min, lat_max = earthquakes[key]['lat_range']
            lon_min, lon_max = earthquakes[key]['lon_range']

            lat_index = np.where((lat_min < lat) & (lat < lat_max))[0]
            lon_index = np.where((lon_min < lon) & (lon < lon_max))[0]

            teq, tau = earthquakes[key]['teq'], earthquakes[key]['tau']

            if len(earthquakes[key]['teq']) == 1:
                self.__analyse_for_once(gqij, lon_index, lat_index, teq, tau)

            elif len(earthquakes[key]['teq']) == 2:
                self.__analyse_for_twice(gqij, lon_index, lat_index, teq, tau)

            else:
                assert False

    def __analyse_for_once(self, gqij, lon_index, lat_index, teq, tau):
        teq, tau = teq[0], tau[0]

        def fit_function1(x, c):
            return c

        def fit_function2(x, c, p):
            return c + p * (1 - np.exp(-(x - teq) / tau))

        year_fractions = self.__get_year_fractions()

        for i in lon_index:
            for j in lat_index:

                sigs = []
                for t in range(len(year_fractions)):
                    sigs.append(gqij[t, j, i])
                sigs = np.array(sigs)

                times_before = []
                sigs_before = []
                times_after = []
                sigs_after = []
                for t in range(len(year_fractions)):
                    if year_fractions[t] < teq:
                        times_before.append(year_fractions[t])
                        sigs_before.append(sigs[t])
                    else:
                        times_after.append(year_fractions[t])
                        sigs_after.append(sigs[t])

                times_before = np.array(times_before)
                sigs_before = np.array(sigs_before)
                times_after = np.array(times_after)
                sigs_after = np.array(sigs_after)

                z1 = MathTool.curve_fit(fit_function1, times_before, sigs_before)[0]
                z2 = MathTool.curve_fit(fit_function2, times_after, sigs_after)[0]
                c1 = z1[0, 0]
                c2, p = z2[0, 0], z2[0, 1]
                seismic = np.zeros_like(year_fractions)
                for t in range(len(year_fractions)):
                    if year_fractions[t] < teq:
                        seismic[t] = c1
                    else:
                        seismic[t] = c2 + p * (1 - np.exp(-(year_fractions[t] - teq) / tau))

                gqij[:, j, i] -= seismic

    def __analyse_for_twice(self, gqij, lon_index, lat_index, teq, tau):

        def fit_function2(x, c, p):
            return c + p * (1 - np.exp(-(x - teq1) / tau1))

        def fit_function3(x, c, p):
            return c + p * (1 - np.exp(-(x - teq2) / tau2))

        teq1, teq2 = teq[0], teq[1]
        tau1, tau2 = tau[0], tau[1]
        time_points = self.__get_year_fractions()

        for i in lon_index:
            for j in lat_index:

                sigs = []
                for t in range(len(time_points)):
                    sigs.append(gqij[t, j, i])
                sigs = np.array(sigs)

                times_before = []
                sigs_before = []
                times_middle = []
                sigs_middle = []
                times_after = []
                sigs_after = []
                for t in range(len(time_points)):
                    if time_points[t] < teq1:
                        times_before.append(time_points[t])
                        sigs_before.append(sigs[t])
                    elif teq1 <= time_points[t] < teq2:
                        times_middle.append(time_points[t])
                        sigs_middle.append(sigs[t])
                    else:
                        times_after.append(time_points[t])
                        sigs_after.append(sigs[t])

                times_before = np.array(times_before)
                sigs_before = np.array(sigs_before)
                times_middle = np.array(times_middle)
                sigs_middle = np.array(sigs_middle)
                times_after = np.array(times_after)
                sigs_after = np.array(sigs_after)

                def fit_function1(x, c):
                    return c

                z1 = MathTool.curve_fit(fit_function1, times_before, sigs_before)[0]
                z2 = MathTool.curve_fit(fit_function2, times_middle, sigs_middle)[0]
                z3 = MathTool.curve_fit(fit_function3, times_after, sigs_after)[0]
                c1 = z1[0, 0]
                c2, p1 = z2[0, 0], z2[0, 1]
                c3, p2 = z3[0, 0], z3[0, 1]
                seismic = np.zeros_like(time_points)
                for t in range(len(time_points)):
                    if time_points[t] < teq1:
                        seismic[t] = c1
                    elif teq1 <= time_points[t] < teq2:
                        seismic[t] = c2 + p1 * (1 - np.exp(-(time_points[t] - teq1) / tau1))
                    else:
                        seismic[t] = c3 + p2 * (1 - np.exp(-(time_points[t] - teq2) / tau2))

                gqij[:, j, i] -= seismic
                pass

        pass



import numpy as np

from sagea.pysrc.auxiliary.MathTool import MathTool
from sagea.pysrc.post_processing.filter.Base import get_poly_func, SHCFilter


class PnMmConfig:
    def __init__(self):
        self.poly_n = 3
        self.start_m = 10

    def set_n(self, n):
        self.poly_n = n

        return self

    def set_m(self, m):
        self.start_m = m

        return self


class PnMm(SHCFilter):
    def __init__(self):
        self.configuration = PnMmConfig()

    def _apply_to_cqlm(self, cqlm: np.ndarray):
        lmax = np.shape(cqlm)[1] - 1
        fit_function = get_poly_func(self.configuration.poly_n)

        def get_fit_array(array):
            t = np.arange(np.shape(array)[0]) + 1
            A = MathTool.get_design_matrix(fit_function, t)
            fit_params = np.linalg.pinv(A) @ array
            t_expand = np.array([t ** p for p in range(self.configuration.poly_n + 1)]).T
            fit_array = t_expand @ fit_params

            return fit_array

        for m in range(self.configuration.start_m, lmax + 1, 1):
            l1 = np.arange(m, lmax + 1, 2)
            l2 = np.arange(m + 1, lmax + 1, 2)
            if len(l1) <= self.configuration.poly_n or len(l2) <= self.configuration.poly_n:
                continue

            this_array1 = cqlm[:, l1, m].T
            this_array2 = cqlm[:, l2, m].T

            fit_this_array1 = get_fit_array(this_array1)
            fit_this_array2 = get_fit_array(this_array2)

            this_array1 -= fit_this_array1
            this_array2 -= fit_this_array2
            cqlm[:, l1, m] = this_array1.T
            cqlm[:, l2, m] = this_array2.T

        return cqlm

    # def apply_to(self, shc: SHC):
    #     cqlm, sqlm = shc.get_cs2d()
    #
    #     length_of_cqlm = np.shape(cqlm)[0]
    #     csqlm = np.concatenate([cqlm, sqlm])
    #     csqlm = self._apply_to_cqlm(csqlm)
    #
    #     cqlm_filtered = csqlm[:length_of_cqlm]
    #     sqlm_filtered = csqlm[length_of_cqlm:]
    #     return SHC(cqlm_filtered, sqlm_filtered)

    def apply_to(self, cqlm, sqlm):

        cqlm, sqlm, single = self._cs_to_3d_array(cqlm, sqlm)

        length_of_cqlm = np.shape(cqlm)[0]
        csqlm = np.concatenate([cqlm, sqlm])
        csqlm = self._apply_to_cqlm(csqlm)

        cqlm_f = csqlm[:length_of_cqlm]
        sqlm_f = csqlm[length_of_cqlm:]

        if single:
            assert cqlm_f.shape[0] == sqlm_f.shape[0] == 1
            return cqlm_f[0], sqlm_f[0]
        else:
            return cqlm_f, sqlm_f



import copy

import numpy as np
from tqdm import trange

from sagea.pysrc.auxiliary.MathTool import MathTool
from sagea.pysrc.post_processing.filter.Base import SHCFilter
from sagea.pysrc.post_processing.harmonic.Harmonic import Harmonic
from sagea.pysrc.post_processing.leakage.Base import Leakage


def keep_signals_in_basin(signals, basin, basin_to_maintain_global_conservation):
    new_signals = signals * basin
    total_signal = MathTool.global_integral(new_signals)

    basin_fixed = total_signal / MathTool.get_acreage(basin_to_maintain_global_conservation)

    new_signals -= np.einsum('i,jk->ijk', basin_fixed, basin_to_maintain_global_conservation)

    return new_signals


class ForwardModelingConfig:
    def __init__(self):
        self.__acceleration_factor = 1
        self.__max_iteration = 50
        self.__basin_to_maintain_global_conservation = None

        self.__basin = None
        self.__harmonic = None
        self.__filter = None

        self.__initial_grid = None
        self.__observed_gqij = None

        self.__log = False

    def set_acceleration_factor(self, factor):
        self.__acceleration_factor = factor
        return self

    def get_acceleration_factor(self):
        return self.__acceleration_factor

    def set_max_iteration(self, max_iteration):
        self.__max_iteration = max_iteration
        return self

    def get_max_iteration(self):
        return self.__max_iteration

    def set_observed_grid(self, gqij: np.ndarray):
        self.__observed_gqij = gqij
        return self

    def get_observed_grid(self):
        return self.__observed_gqij

    def set_filter(self, cs_filter: SHCFilter):
        self.__filter = cs_filter
        return self

    def get_filter(self):
        return self.__filter

    def set_harmonic(self, harmonic: Harmonic):
        self.__harmonic = harmonic
        return self

    def get_harmonic(self):
        return self.__harmonic

    def set_basin_conservation(self, basin: np.ndarray):
        self.__basin_to_maintain_global_conservation = basin

        return self

    def get_basin_conservation(self):
        return self.__basin_to_maintain_global_conservation

    def set_basin(self, basin: np.ndarray):
        self.__basin = basin

        return self

    def get_basin(self):
        return self.__basin

    def set_print_log(self, log=True):
        self.__log = log
        return self

    def get_print_log(self):
        return self.__log


class ForwardModeling(Leakage):
    def __init__(self):
        super().__init__()

        self.configuration = ForwardModelingConfig()

    def apply_to(self, gqij: np.ndarray, get_grid=False):
        if self.configuration.get_observed_grid() is not None:
            observed_model = self.configuration.get_observed_grid()
        else:
            observed_model = copy.deepcopy(gqij)

        basin = self.configuration.get_basin()

        basin_to_conservation = self.configuration.get_basin_conservation()
        shc_filter = self.configuration.get_filter()
        har = self.configuration.get_harmonic()
        acceleration_factor = self.configuration.get_acceleration_factor()
        max_iter_times = self.configuration.get_max_iteration()

        true_model = keep_signals_in_basin(gqij, basin, basin_to_conservation)
        iter_times = 0
        print_log = self.configuration.get_print_log()

        if print_log:
            ran = trange(0, max_iter_times, 1)
        else:
            ran = range(0, max_iter_times, 1)

        for iter_times in ran:
            cqlm, sqlm = har.analysis(true_model)
            cqlm_filtered, sqlm_filtered = shc_filter.apply_to(cqlm, sqlm)

            grids_predicted = har.synthesis(cqlm_filtered, sqlm_filtered)

            grids_difference = (observed_model - grids_predicted) * basin

            true_model += grids_difference * acceleration_factor
            true_model = keep_signals_in_basin(true_model, basin, basin_to_conservation)

        print()

        if get_grid:
            return true_model

        else:
            return MathTool.global_integral(true_model * basin)

    def format(self):
        return 'Forward modeling'



import numpy as np
from tqdm import trange

from sagea.pysrc.post_processing.geometric_correction.old.GC import GeometricalCorrection as GC
from sagea.pysrc.post_processing.geometric_correction.old.GeoMathKit import GeoMathKit
from sagea.pysrc.post_processing.geometric_correction.old.Setting import Assumption, FieldType

import sagea.pysrc.auxiliary.Preference as Enums


class GeometricalCorrectionConfig:
    def __init__(self):
        self.__assumption = Assumption.ActualEarth
        self.__kind = FieldType.EWH
        self.__lmax = None

        self.__lat = None
        self.__lon = None

        pass

    def set_assumption(self, assumption: Assumption):
        self.__assumption = assumption
        return self

    def get_assumption(self):
        return self.__assumption

    def set_kind(self, kind: FieldType):
        self.__kind = kind
        return self

    def get_kind(self):
        return self.__kind

    def set_lmax(self, lmax):
        self.__lmax = lmax
        return self

    def get_lmax(self):
        return self.__lmax

    def set_lat_lon(self, lat, lon):
        self.__lat, self.__lon = lat, lon
        return self

    def get_lat_lon(self):
        return self.__lat, self.__lon

    def get_lat(self):
        return self.__lat

    def get_lon(self):
        return self.__lon


class GeometricalCorrection:
    def __init__(self):
        self.configuration = GeometricalCorrectionConfig()

    def apply_to(self, cqlm, sqlm, assumption, log=False):
        assert assumption in Enums.GeometricCorrectionAssumption

        if assumption == Enums.GeometricCorrectionAssumption.Sphere:
            assumption = Assumption.Sphere
        elif assumption == Enums.GeometricCorrectionAssumption.Ellipsoid:
            assumption = Assumption.Ellipsoid
        elif assumption == Enums.GeometricCorrectionAssumption.ActualEarth:
            assumption = Assumption.ActualEarth
        else:
            assert False

        grid_space = 0.5
        lat = np.arange(-90, 90 + grid_space / 2, grid_space)
        lon = np.arange(-180 + grid_space / 2, 180 + grid_space / 2, grid_space)
        lmax = np.shape(cqlm)[1] - 1

        self.configuration.set_lat_lon(lat, lon)
        self.configuration.set_lmax(lmax)
        self.configuration.set_assumption(assumption)

        cqlm_new, sqlm_new = [], []

        if log:
            ran = trange(len(cqlm), desc="geometric correction")
        else:
            ran = range(len(cqlm))

        for i in ran:
            clm, slm = cqlm[i], sqlm[i]

            gc = GC().configure(
                Nmax=self.configuration.get_lmax(),
                lat=self.configuration.get_lat(),
                lon=self.configuration.get_lon(),
                assumption=self.configuration.get_assumption(),
                kind=self.configuration.get_kind()
            )

            c1d = GeoMathKit.CS_2dTo1d(clm)
            s1d = GeoMathKit.CS_2dTo1d(slm)

            cs1d_corrected = gc.setInput(GravityField=(c1d, s1d)).correct()
            clm_corrected, slm_corrected = GeoMathKit.CS_1dTo2d(cs1d_corrected[0]), GeoMathKit.CS_1dTo2d(
                cs1d_corrected[1])

            cqlm_new.append(clm_corrected)
            sqlm_new.append(slm_corrected)

        if log:
            print(f"done!")

        cqlm_new = np.array(cqlm_new)
        sqlm_new = np.array(sqlm_new)

        return cqlm_new, sqlm_new



#!/use/bin/env python
# coding=utf-8
# @Author  : Shuhao Liu
# @Time    : 2025/6/13 10:34 
# @File    : __SHC__.py
import copy

import numpy as np

from sagea.pysrc.auxiliary.Preference import GeoConstants, SHNormalization, Dimension
import sagea.pysrc.auxiliary.Preference as Enums
from sagea.pysrc.auxiliary.TimeTool import TimeTool
from sagea.pysrc.auxiliary.MathTool import MathTool
from sagea.pysrc.post_processing.filter.GetSHCFilter import get_filter
from sagea.pysrc.auxiliary import Preference
from sagea.pysrc.auxiliary.LoveNumber import LoveNumber
from sagea.pysrc.post_processing.convert_field_physical_quantity.ConvertSHC import ConvertSHC
from sagea.pysrc.post_processing.geometric_correction.GeometricalCorrection import GeometricalCorrection
from sagea.pysrc.post_processing.harmonic.Harmonic import Harmonic
from sagea.pysrc.post_processing.replace_low_deg.ReplaceLowDegree import ReplaceLowDegree


class SHC:
    """
    This class is to store the spherical harmonic coefficients (SHCs) for the use in necessary data processing.

    Attribute self.value stores the coefficients in 2d-array (in numpy.ndarray) combined with c and s.
    which are sorted by degree, for example,
    numpy.ndarray: [[c1[0,0]; s1[1,1], c1[1,0], c1[1,1]; s1[2,2], s1[2,1], c1[2,0], c1[2,1], c1[2,2]; ...],
     [c2[0,0]; s2[1,1], c2[1,0], c2[1,1]; s2[2,2], s2[2,1], c2[2,0], c2[2,1], c2[2,2]; ...],
     [                                        ...                                         ]].
    Note that even it stores only one set of SHCs, the array is still 2-dimension, i.e.,
    [[c1[0,0]; s1[1,1], c1[1,0], c1[1,1]; s1[2,2], s1[2,1], c1[2,0], c1[2,1], c1[2,2]; ...]].

    Attribute self.normalization indicates the normalization of the SHCs (in sagea.Preference.SHNormalization), for example,
    sagea.Preference.SHNormalization.full.

    Attribute self.physical_dimension indicates the physical dimension of the SHCs (in sagea.Preference.PhysicalDimensions).
    """

    def __init__(self, c, s=None, normalization=None, dimension: Preference.Dimension = None):
        """

        :param c:
                if s is not None:
                    harmonic coefficients c in 2-dimension (l,m), or a series (q,l,m);
                else:
                    1-dimension array sorted by degree [c00, s11, c10, c11, s22, s21, ...]
                    or 1-dimension array sorted by degree [[c00, s11, c10, c11, s22, s21, ...],[...]]

        :param s: harmonic coefficients s in 2-dimension (l,m), or a series (q,l,m),
                or None.

        :param normalization: in Preference.SHNormalization, defaults to .full.

        :param dimension: in Preference.PhysicalDimensions, defaults to .Dimensionless.
        """
        if s is None:
            self.value = np.array(c)
            if len(self.value.shape) == 1:
                self.value = np.array([self.value])

        else:
            assert np.shape(c) == np.shape(s)

            if len(np.shape(c)) == 2:
                self.value = MathTool.cs_combine_to_triangle_1d(c, s)

            elif len(np.shape(c)) == 3:
                cs = []
                for i in range(np.shape(c)[0]):
                    this_cs = MathTool.cs_combine_to_triangle_1d(c[i], s[i])
                    cs.append(this_cs)
                self.value = np.array(cs)

        if len(np.shape(self.value)) == 1:
            self.value = self.value[None, :]

        assert len(np.shape(self.value)) == 2

        if normalization is None:
            normalization = Preference.SHNormalization.full
        assert normalization in Preference.SHNormalization
        self.__normalization = normalization

        if dimension is None:
            dimension = Preference.Dimension.Dimensionless
        assert dimension in Preference.Dimension

        self.__dimension = dimension

    def append(self, *params):
        """

        :param params: One parameter of instantiated SHC,
         or two parameters of c and s with the same input requirement as SHC.
        :return: self
        """
        assert len(params) in (1, 2)

        if len(params) == 1:
            if issubclass(type(params[0]), SHC):
                shc = params[0]
            else:
                shc = SHC(params[0])

        else:
            shc = SHC(*params)

        assert np.shape(shc.value)[-1] == np.shape(self.value)[-1]

        self.value = np.concatenate([self.value, shc.value])

        return self

    def is_series(self):
        """
        To determine whether the spherical harmonic coefficients stored in this class are one group or multiple groups.
        :return: bool, True if it stores multiple groups, False if it stores only one group.
        """
        return self.get_length() != 1

    def get_length(self):
        """
        To get the number of sets.
        """
        return np.shape(self.value)[0]

    def __len__(self):
        return self.get_length()

    def get_lmax(self):
        """

        :return: int, max degree/order of the spherical harmonic coefficients stored in this class.
        """
        length_of_cs1d = np.shape(self.value)[-1]
        lmax = int(np.sqrt(length_of_cs1d) - 1)

        return lmax

    def get_physical_dimension(self):
        """

        :return:
        """

        return self.__dimension

    def set_physical_dimension(self, physical_dimension: Dimension):
        self.__dimension = physical_dimension

        return self

    def get_normalization(self):
        """

        :return:
        """

        return self.__normalization

    def set_normalization(self, normalization: SHNormalization):
        self.__normalization = normalization
        return self

    def get_cs2d(self, fill_value=0):
        """
        return: cqlm, sqlm. Both cqlm and sqlm are 3-dimension, EVEN IF NOT self.is_series()
        """
        lmax = self.get_lmax()

        num_of_series = np.shape(self.value)[0]

        cqlm = np.zeros((num_of_series, lmax + 1, lmax + 1))
        sqlm = np.zeros((num_of_series, lmax + 1, lmax + 1))

        for i in range(num_of_series):
            this_cs = self.value[i]
            this_clm, this_slm = MathTool.cs_decompose_triangle1d_to_cs2d(this_cs)
            cqlm[i, :, :] = this_clm
            sqlm[i, :, :] = this_slm

        return cqlm, sqlm

    def __de_average(self):
        if self.is_series():
            self.value -= np.mean(self.value, axis=0)
        else:
            raise Exception

    def get_average(self):
        return SHC(np.mean(self.value, axis=0))

    def de_background(self, background=None):
        """
        if background is None, de average
        """
        if background is None:
            self.__de_average()

        else:
            assert isinstance(background, SHC)
            assert not background.is_series()

            self.value -= background.value

    @staticmethod
    def identity(lmax: int):
        basis_num = (lmax + 1) ** 2
        cs = np.eye(basis_num)

        return SHC(cs)

    def __add__(self, other):
        assert isinstance(other, SHC)
        assert self.get_lmax() == other.get_lmax()

        return SHC(self.value + other.value)

    def __sub__(self, other):
        assert isinstance(other, SHC)
        assert self.get_lmax() == other.get_lmax()

        return SHC(self.value - other.value)

    def add(self, shc, lbegin=None, lend=None):
        if (lend is None) and (lbegin is None):
            self.value += shc.value
            return self

        else:
            assert (lend is None) or (0 <= lend <= self.get_lmax())
            assert (lbegin is None) or (0 <= lbegin <= self.get_lmax())

            shc_copy = copy.deepcopy(shc)

            if lend is not None:
                shc_copy.value[:, (lend + 1) ** 2:] = 0
            if lbegin is not None:
                shc_copy.value[:, :lbegin ** 2] = 0

            return self.add(shc_copy)

    def subtract(self, shc, lbegin=None, lend=None):
        if (lend is None) and (lbegin is None):
            self.value -= shc.value
            return self

        else:
            assert (lend is None) or (0 <= lend <= self.get_lmax())
            assert (lbegin is None) or (0 <= lbegin <= self.get_lmax())

            shc_copy = copy.deepcopy(shc)

            if lend is not None:
                shc_copy.value[:, (lend + 1) ** 2:] = 0
            if lbegin is not None:
                shc_copy.value[:, :lbegin ** 2] = 0

            return self.subtract(shc_copy)

    def get_degree_rms(self):
        cqlm, sqlm = self.get_cs2d()
        return MathTool.cs_get_degree_rms(cqlm, sqlm)

    def get_degree_rss(self):
        cqlm, sqlm = self.get_cs2d()
        return MathTool.cs_get_degree_rss(cqlm, sqlm)

    def get_cumulative_degree_rss(self):
        cqlm, sqlm = self.get_cs2d()
        return MathTool.cs_get_cumulative_degree_rss(cqlm, sqlm)

    def get_std(self):
        cs_std = np.std(self.value, axis=0)
        return SHC(cs_std)

    def convert_type(self, to_type: Preference.Dimension):
        from_type = self.__dimension

        assert from_type in Preference.Dimension and from_type.value <= 99

        assert to_type in Preference.Dimension and to_type.value <= 99

        lmax = self.get_lmax()
        LN = LoveNumber()
        LN.configuration.set_lmax(lmax)

        ln = LN.get_Love_number()

        convert = ConvertSHC()
        convert.configuration.set_Love_number(ln).set_input_type(from_type).set_output_type(to_type)

        self.value = convert.apply_to(self.value)
        self.__dimension = to_type
        return self

    def filter(self, method: Preference.SHCFilterType or Preference.SHCDecorrelationType, param: tuple = None):
        assert method in Enums.SHCFilterType or method in Preference.SHCDecorrelationType
        cqlm, sqlm = self.get_cs2d()
        filtering = get_filter(method, param, lmax=self.get_lmax())
        cqlm_f, sqlm_f = filtering.apply_to(cqlm, sqlm)
        self.value = MathTool.cs_combine_to_triangle_1d(cqlm_f, sqlm_f)

        return self

    def to_GRD(self, grid_space=None):
        from sagea.pysrc.data_class.__GRD__ import GRD

        """pure synthesis"""

        if grid_space is None:
            grid_space = int(180 / self.get_lmax())

        lat, lon = MathTool.get_global_lat_lon_range(grid_space)

        lmax = self.get_lmax()
        har = Harmonic(lat, lon, lmax, option=1)

        cqlm, sqlm = self.get_cs2d()

        if self.__dimension in (Preference.Dimension.HorizontalDisplacementEast,
                                Preference.Dimension.HorizontalDisplacementNorth):
            grid_data = har.synthesis(cqlm, sqlm, special_type=self.__dimension)

        else:
            grid_data = har.synthesis(cqlm, sqlm)

        grid = GRD(grid_data, lat, lon, option=1, dimension=self.__dimension)
        if grid.get_dimension() == Preference.Dimension.DimensionlessMask:
            grid.limiter(0.5, 1, 0)

        return grid

    def geometric(self, assumption: Preference.GeometricCorrectionAssumption, log=False):
        gc = GeometricalCorrection()
        cqlm, sqlm = self.get_cs2d()
        cqlm_new, sqlm_new = gc.apply_to(cqlm, sqlm, assumption=assumption, log=log)
        self.value = MathTool.cs_combine_to_triangle_1d(cqlm_new, sqlm_new)

        return self

    def replace_low_degs(self, dates_begin, dates_end, low_deg: dict,
                         c10=False, c11=False, s11=False, c20=False, c30=False):
        assert False, "replace_low_degs() is not implemented, please use replace()"

        assert len(dates_begin) == len(dates_end) == len(self.value)

        replace_or_not = (c10, c11, s11, c20, c30)
        low_ids = ("c10", "c11", "s11", "c20", "c30")
        for i in range(len(low_ids)):
            if replace_or_not[i]:
                assert low_ids[i] in low_deg.keys(), f"input low_deg should include key {low_ids[i]}"

        replace_low_degs = ReplaceLowDegree()
        replace_low_degs.configuration.set_replace_c10(c10).set_replace_c11(c11).set_replace_s11(s11)
        replace_low_degs.configuration.set_replace_c20(c20).set_replace_c30(c30)

        replace_low_degs.set_low_degrees(low_deg)
        cqlm, sqlm = replace_low_degs.apply_to(*self.get_cs2d(), begin_dates=dates_begin, end_dates=dates_end)
        self.value = MathTool.cs_combine_to_triangle_1d(cqlm, sqlm)

        return self

    def replace(self, index: str, new: np.ndarray):
        """

        :param index: str, like "c2,0" or "s1,1"
        :param new:
        :return:
        """
        assert isinstance(new, SHC) or isinstance(new, np.ndarray)

        index1d = MathTool.get_cs_1d_index(index)

        if isinstance(new, SHC):
            new_array = new.value[:, index1d]
        else:
            new_array = new

        assert len(new_array) == len(self)

        index_new_array_valid = np.where(new_array == new_array)

        self.value[index_new_array_valid, index1d] = new_array[index_new_array_valid]

    def linear_expand(self, time):
        assert not self.is_series()

        year_frac = TimeTool.convert_date_format(time,
                                                 input_type=TimeTool.DateFormat.ClassDate,
                                                 output_type=TimeTool.DateFormat.YearFraction)

        year_frac = np.array(year_frac)

        trend = self.value[0]
        value = year_frac[:, None] @ trend[None, :]
        return SHC(value)

    def regional_extraction(self, shc_region, average=True):
        assert isinstance(shc_region, SHC)

        extraction = (shc_region.value @ self.value.T) * (GeoConstants.radius_earth ** 2)

        if average:
            extraction /= (shc_region.value[:, None, 0] * (GeoConstants.radius_earth ** 2))

        else:
            if self.__normalization == SHNormalization.full:
                extraction *= (4 * np.pi)
            else:
                assert False

        return extraction



from abc import ABC, abstractmethod

from sagea.pysrc.post_processing.filter.Base import SHCFilter
from sagea.pysrc.post_processing.harmonic.Harmonic import Harmonic


class Leakage(ABC):
    def __init__(self):
        pass

    @abstractmethod
    def apply_to(self, gqij, get_grid=False):
        pass

    @abstractmethod
    def format(self):
        pass


def filter_grids(gqij, shc_filter: SHCFilter, harmonic: Harmonic):
    cqlm, sqlm = harmonic.analysis(gqij)

    cqlm_f, sqlm_f = shc_filter.apply_to(cqlm, sqlm)
    gqij_filtered = harmonic.synthesis(cqlm_f, sqlm_f)

    return gqij_filtered



#!/usr/bin/env python
# -*- coding:utf-8 -*-
# author:Yang Fan
# mailbox: yfan_cge@hust.edu.cn
# address: Huazhong University of Science and Technology, Wuhan, China
# datetime:2020/6/15 下午4:41
# software: Atmosphere de-aliasing modelling
# usage of this file: calculate the geoid undulation
import pathlib

import numpy as np

from sagea.pysrc.post_processing.geometric_correction.old.Harmonic import Harmonic
from sagea.pysrc.post_processing.geometric_correction.old.LoadSH import LoadSH, Gif48
from sagea.pysrc.post_processing.geometric_correction.old.LoveNumber import LoveNumber
from sagea.pysrc.post_processing.geometric_correction.old.RefEllipsoid import EllipsoidType, RefEllipsoid
from sagea.pysrc.post_processing.geometric_correction.old.Setting import SynthesisType, LoveNumberType


class GeoidUndulation:
    """
    Definitions and calculations refer to the
    http://icgem.gfz-potsdam.de/faq

    Notice: A more detailed and more accurate computation refer to:
    https://www.sciencedirect.com/science/article/pii/B9780444527486001565
    https://shtools.github.io/SHTOOLS/pymakegeoidgriddh.html
    """

    __LN = LoveNumber('../data/')

    def __init__(self, ell: EllipsoidType):
        self.__SH = None
        self.__Nmax = None

        self.__default()

        self.__ellipsoid = RefEllipsoid(ell)
        self.__Har = Harmonic(GeoidUndulation.__LN, Parallel=-1).setEllipsoid(ell).setLoveNumMethod(LoveNumberType.Wang)

        pass

    def setGravityModel(self, SH: LoadSH, Nmax: int):
        """

        :param SH: gravity model
        :param Nmax: max degree
        :return:
        """
        self.__SH = SH
        self.__Nmax = Nmax

        return self

    def getGeoid(self, lat, lon):
        """
        get geoid undulation of N*M points
        :param lat: geocentric latitude in degree, [dimension N]
        :param lon: geocentric longitude in degree, [dimension M]
        :return: [dimension N * M]
        """
        C, S = self.__SH.getCS(self.__Nmax)
        # C, S = np.ones(81*161), np.ones(81*161)
        C[0:3] = 0.0
        S[0:3] = 0.0

        C_normal = self.__ellipsoid.NormalGravity

        ll = len(C_normal)
        C[0:ll] -= C_normal

        geoid = self.__Har.synthesis(C, S, self.__Nmax, lat, lon, SynthesisType.Geoidheight)

        return geoid

    def __default(self):
        """
        a default gravity model "gif48" is configured to obtain the geoid undulation.
        :return:
        """
        self.__Nmax = 160
        self.__SH = Gif48().load(str(pathlib.Path(__file__).parent / "../../../../__data__/auxiliary/GIF48.gfc"))
        pass


def demo1():
    lat = np.arange(-89.75, 90, 0.5)
    lon = np.arange(0.25, 360, 0.5)

    geoid = GeoidUndulation(EllipsoidType.GRS80_IERS2010).getGeoid(lat, lon)

    pass


if __name__ == '__main__':
    # demo1()
    demo1()



import numpy as np

from pysrc.auxiliary.aux_tool.MathTool import MathTool


class ExtractSpatialSignalConfig:
    def __init__(self):
        lat_range = np.arange(-89.5, 90.5, 1)
        lon_range = np.arange(-179.5, 180.5, 1)

        self.lat_range, self.lon_range = MathTool.get_colat_lon_rad(lat_range, lon_range)

    def set_lat_lon_range(self, lat, lon, option=1):
        """
        :param lat: geographic latitude in [deg] if option = 1 or co-latitude in [rad]
        :param lon: geographic longitude in [deg] if option = 1 or longitude in [rad]
        :param option: 1 or 0
        """
        assert option in (1, 0)
        if option == 1:
            lat, lon = MathTool.get_colat_lon_rad(lat, lon)

        self.lat_range = lat
        self.lon_range = lon

        return self



import numpy as np

from sagea.pysrc.auxiliary.Preference import GeoConstants

from sagea.pysrc.auxiliary.MathTool import MathTool
from sagea.pysrc.post_processing.filter.Base import get_gaussian_weight_1d, SHCFilter


class FanConfig:
    def __init__(self):
        """default setting"""
        self.filtering_radius1 = 500 * 1000  # unit[m]
        self.filtering_radius2 = 500 * 1000  # unit[m]

        self.earth_radius = GeoConstants.radius_earth
        self.lmax = 60

    def set_filtering_params(self, radius1: int, radius2: int):
        """
        filtering radius,
        the distance from the center of the kernel to the point where the filtering kernel decays to its half.
        :param radius1: int, unit [km]
        :param radius2: int, unit [km]
        :return:
        """
        self.filtering_radius1 = radius1 * 1000
        self.filtering_radius2 = radius2 * 1000
        return self

    def set_lmax(self, lmax: int):
        """
        max degree/order
        :param lmax: int
        :return:
        """
        self.lmax = lmax
        return self

    def __str__(self):
        """

        :return: str, information of the configuration
        """

        info = f'Filtering Radius 1:\t{self.filtering_radius1 / 1000} km\n' \
               f'Filtering Radius 2:\t{self.filtering_radius2 / 1000} km\n' \
               f'Earth Radius:\t{self.earth_radius / 1000} km\n' \
               f'Max degree/order:\t{self.lmax}'

        return info


class Fan(SHCFilter):
    def __init__(self):
        self.configuration = FanConfig()

    def config(self, config: FanConfig):
        self.configuration = config

    def __get_weight_matrix(self):
        """
        get C/S filtering weight sorted by degree, that is,
        [[w00,   0,   0, ...]
         [w10, w11,   0, ...]
         [w20, w21, w22, ...]
         [       ...       ]],
        :return:
        """
        lmax = self.configuration.lmax

        w1 = get_gaussian_weight_1d(lmax, self.configuration.filtering_radius1, self.configuration.earth_radius)
        w2 = get_gaussian_weight_1d(lmax, self.configuration.filtering_radius2, self.configuration.earth_radius)

        matrix = w1[:, None] @ w2[None, :]

        return np.tril(matrix)

    def get_weight_cs1d(self):
        """
        get C/S filtering weight sorted by degree, that is,
        [w(s00), w(c10), w(c11), w(s11), w(c20), w(c21), w(s21), w(c22), w(s22), ...]
        :return: 1-d array in length (self.lmax + 1) **2
        """

        weight_mat = self.__get_weight_matrix()
        gs_weight_cs1d = MathTool.cs_combine_to_triangle_1d(weight_mat, weight_mat)

        return gs_weight_cs1d

    def apply_to(self, cqlm, sqlm):

        cqlm, sqlm, single = self._cs_to_3d_array(cqlm, sqlm)

        weight_matrix = self.__get_weight_matrix()
        cqlm_f, sqlm_f = cqlm * weight_matrix, sqlm * weight_matrix

        if single:
            assert cqlm_f.shape[0] == sqlm_f.shape[0] == 1
            return cqlm_f[0], sqlm_f[0]
        else:
            return cqlm_f, sqlm_f



import numpy as np

from sagea.pysrc.auxiliary.MathTool import MathTool
from sagea.pysrc.post_processing.leakage.BaseModelDriven import ModelDriven


class ScalingGrid(ModelDriven):
    def apply_to(self, grids: np.ndarray, get_grid=True):
        grid_factor = self._get_scaling_scale_grid(scale_type=self.configuration.scale_type)

        scaled_grids = grids * grid_factor

        if get_grid:
            return scaled_grids

        else:
            basin_map = self.configuration.basin_map
            return MathTool.global_integral(scaled_grids * basin_map) / MathTool.get_acreage(basin_map)

    def format(self):
        return 'scaling grid'



import re

import numpy as np

from sagea.pysrc.auxiliary.Preference import GeoConstants


class MathTool:
    @staticmethod
    def cs_combine_to_triangle_1d(c: np.ndarray, s: np.ndarray):
        """
        combine the SHC C or S in 2-dimensional matrix to 1-dimension vector,
        or SHC C or S in 3-dimensional matrix to 2-dimension vector.
        Example:

        input
        00
        10 11
        20 21 22
        30 31 32 33

        return cs 1d-array which is formed as
        [c(0,0); s(1,1), c(1,0), c(1,1); s(2,2), s(2,1), c(2,0), c(2,1), c(2,2); s(3,3), s(3,2), s(3,1), c(3,0), ...].
        """
        assert np.shape(c) == np.shape(s)
        lmax = np.shape(c)[-1] - 1

        ones_tril = np.tril(np.ones((lmax + 1, lmax + 1)))
        ones_tri = np.concatenate([ones_tril[:, -1:0:-1], ones_tril], axis=1)
        index_tri = np.where(ones_tri == 1)

        cs_tri = MathTool.cs_combine_to_triangle(c, s)

        if cs_tri.ndim == 2:
            return cs_tri[index_tri]

        elif cs_tri.ndim == 3:
            return np.array([cs_tri[i][index_tri] for i in range(np.shape(cs_tri)[0])])

        else:
            raise Exception

    @staticmethod
    def cs_combine_to_triangle(c: np.ndarray, s: np.ndarray):
        """
        :param c: 2- or 3-d array clm or cqlm
        :param s: 2- or 3-d array slm or sqlm
        return: 2d-array like /s|c\, for example,
        [[0,   ...,   0, c00,   0,  ...,    0],
         [0,   ..., s11, c10, c11,  ...,    0],
         [0,   ..., s21, c20, c21,  ...,    0],
         [...  ...,      ...,       ...,  ...],
         [sii, ..., si1, ci0, ci1,  ...,  cii]]

        or 3d-array with the last two dimensions representing format as above if input are 3-d array.
        """

        assert np.shape(c) == np.shape(s)

        if c.ndim == 2:
            return np.concatenate([s[:, -1:0:-1], c], axis=1)

        elif c.ndim == 3:
            return np.array([np.concatenate([s[i, :, -1:0:-1], c[i]], axis=1) for i in range(np.shape(c)[0])])

        else:
            raise Exception

    @staticmethod
    def cs_decompose_triangle1d_to_cs2d(cs: np.ndarray, fill=0.):
        """
        :param cs: 1d-array sorted as
        [c(0,0); s(1,1), c(1,0), c(1,1); s(2,2), s(2,1), c(2,0), c(2,1), c(2,2); s(3,3), s(3,2), s(3,1), c(3,0), ...],
        or 2d-array as
        [
        [c1(0,0); s1(1,1), ...],
        [c2(0,0); s2(1,1), ...],
        ...
        ]

        :param fill: filled value, defaults to 0.
        return: 3d-array c_qlm, 3-array s_qlm
        """
        assert cs.ndim in (1, 2)

        if cs.ndim == 1:
            length_cs1d = len(cs)
            lmax = int(np.sqrt(length_cs1d) - 1)
            shape2d = (lmax + 1, lmax + 1)

            clm, slm = np.full(shape2d, fill, dtype=np.float32), np.full(shape2d, fill, dtype=np.float32)

            for l in range(lmax + 1):
                for m in range(l + 1):
                    c_index_tri1d = int(l ** 2 + l + m)
                    clm[l, m] = cs[c_index_tri1d]

                    if m > 0:
                        s_index_tri1d = int(l ** 2 + l - m)
                        slm[l, m] = cs[s_index_tri1d]

            return np.array([clm]), np.array([slm])

        else:
            cqlm, sqlm = [], []
            for i in range(np.shape(cs)[0]):
                clm, slm = MathTool.cs_decompose_triangle1d_to_cs2d(cs[i])
                cqlm.append(clm)
                sqlm.append(slm)

            return np.array(cqlm), np.array(sqlm)

    @staticmethod
    def get_cs_1d_index(c: str, l: int = None, m: int = None):
        c = c.lower()
        assert c[0] in ("c", "s")

        if len(c) > 0:
            assert l is None and m is None

            re_pat = r"([cs])(\d+),(\d+)"
            re_match = re.match(re_pat, c)

            assert re_match is not None

            c = re_match.groups()[0]
            l = int(re_match.groups()[1])
            m = int(re_match.groups()[2])

        assert m <= l

        index1d_l0 = l ** 2 + l

        if c == "c":
            index1d_lm = index1d_l0 + m
        elif c == "s":
            index1d_lm = index1d_l0 - m
        else:
            assert False

        return index1d_lm

    @staticmethod
    def cs_get_degree_rms(cqlm, sqlm):
        assert np.shape(cqlm) == np.shape(sqlm)
        if len(np.shape(cqlm)) == 2:
            cqlm = np.array([cqlm])
            sqlm = np.array([sqlm])

        shape = np.shape(cqlm)
        lmax = np.shape(cqlm)[1] - 1

        rms = np.zeros((shape[0], lmax + 1))

        for i in range(lmax + 1):
            rms_this_degree = np.sum(cqlm[:, i, :i + 1] ** 2 + sqlm[:, i, :i + 1] ** 2, axis=1)
            rms_this_degree = np.sqrt(rms_this_degree / ((i + 1) ** 2))
            rms[:, i] = rms_this_degree

        return rms

    @staticmethod
    def cs_get_degree_rss(cqlm, sqlm):
        assert np.shape(cqlm) == np.shape(sqlm)
        if len(np.shape(cqlm)) == 2:
            cqlm = np.array([cqlm])
            sqlm = np.array([sqlm])

        shape = np.shape(cqlm)
        lmax = np.shape(cqlm)[1] - 1

        rss = np.zeros((shape[0], lmax + 1))

        for i in range(lmax + 1):
            rss_this_degree = np.sum(cqlm[:, i, :i + 1] ** 2 + sqlm[:, i, :i + 1] ** 2, axis=1)
            rss_this_degree = np.sqrt(rss_this_degree)
            rss[:, i] = rss_this_degree

        return rss

    @staticmethod
    def cs_get_cumulative_degree_rss(cqlm, sqlm):
        assert np.shape(cqlm) == np.shape(sqlm)
        rss = MathTool.cs_get_degree_rss(cqlm, sqlm)
        crss = np.zeros_like(rss)
        for i in range(rss.shape[1]):
            crss[:, i] = np.sqrt(np.sum(rss[:, :i + 1] ** 2, axis=1))

        return crss

    @staticmethod
    def get_global_lat_lon_range(resolution):
        """
        get geophysical latitude and longitude range with a given spatial resolution, i.e., the grid space.
        :param resolution: in unit [degree]
        :return: tuple with two elements of 1d-array, latitude and longitude range in unit [degree]
        """
        lat = np.arange(-90 + resolution / 2, 90 + resolution / 2, resolution)
        lon = np.arange(-180 + resolution / 2, 180 + resolution / 2, resolution)

        return lat, lon

    @staticmethod
    def get_Legendre(lat, lmax: int, option=0):
        """
        get associated legendre function up to degree/order lmax in Lat. Be careful that this applies the 4-pi normalizaton.
        :param lat: ndarray, co-latitude if option=0, unit[rad]; geophysical latitude if option = others, unit[degree]
        :param lmax: int, max degree
        :param option:
        :return: 3d-ndarray, indexes stand for (co-lat[rad], degree l, order m)
        """

        if option != 0:
            lat = (90. - lat) / 180. * np.pi

        if type(lat) is np.ndarray:
            lsize = np.size(lat)
        else:
            lsize = 1

        pilm = np.zeros((lsize, lmax + 1, lmax + 1))
        pilm[:, 0, 0] = 1
        pilm[:, 1, 1] = np.sqrt(3) * np.sin(lat)

        '''For the diagonal element'''
        for n in range(2, lmax + 1):
            pilm[:, n, n] = np.sqrt((2 * n + 1) / (2 * n)) * np.sin(lat) * pilm[:, n - 1, n - 1]

        for n in range(1, lmax + 1):
            pilm[:, n, n - 1] = np.sqrt(2 * n + 1) * np.cos(lat) * pilm[:, n - 1, n - 1]

        for n in range(2, lmax + 1):
            for m in range(n - 2, -1, -1):
                pilm[:, n, m] = \
                    np.sqrt((2 * n + 1) / ((n - m) * (n + m)) * (2 * n - 1)) \
                    * np.cos(lat) * pilm[:, n - 1, m] \
                    - np.sqrt((2 * n + 1) / ((n - m) * (n + m)) * (n - m - 1) * (n + m - 1) / (2 * n - 3)) \
                    * pilm[:, n - 2, m]

        return pilm

    @staticmethod
    def get_Legendre_1d_index(lat, Nmax: int, option=0):
        """
        get legendre function up to degree/order Nmax in Lat.
        :param lat: Co-latitude if option=0, unit[rad]; geophysical latitude if option = others, unit[degree]
        :param Nmax:
        :param option:
        :return:
        """

        if option != 0:
            lat = (90. - lat) / 180. * np.pi

        NMmax = int((Nmax + 1) * (Nmax + 2) / 2)

        if type(lat) is np.ndarray:
            Nsize = np.size(lat)
        else:
            Nsize = 1

        Pnm = np.zeros((NMmax, Nsize))

        Pnm[MathTool.getIndex(0, 0)] = 1

        Pnm[MathTool.getIndex(1, 1)] = np.sqrt(3) * np.sin(lat)

        '''For the diagonal element'''
        for n in range(2, Nmax + 1):
            Pnm[MathTool.getIndex(n, n)] = np.sqrt((2 * n + 1) / (2 * n)) * np.sin(lat) * Pnm[
                MathTool.getIndex(n - 1, n - 1)]

        for n in range(1, Nmax + 1):
            Pnm[MathTool.getIndex(n, n - 1)] = np.sqrt(2 * n + 1) * np.cos(lat) * Pnm[
                MathTool.getIndex(n - 1, n - 1)]

        for n in range(2, Nmax + 1):
            for m in range(n - 2, -1, -1):
                Pnm[MathTool.getIndex(n, m)] = \
                    np.sqrt((2 * n + 1) / ((n - m) * (n + m)) * (2 * n - 1)) \
                    * np.cos(lat) * Pnm[MathTool.getIndex(n - 1, m)] \
                    - np.sqrt((2 * n + 1) / ((n - m) * (n + m)) * (n - m - 1) * (n + m - 1) / (2 * n - 3)) \
                    * Pnm[MathTool.getIndex(n - 2, m)]

        return Pnm

    @staticmethod
    def get_Legendre_derivative(lat, lmax: int, option=0):
        """
        get derivative of legendre function up to degree/order lmax in Lat.
        :param lat: ndarray, co-latitude if option=0, unit[rad]; geophysical latitude if option = others, unit[degree]
        :param lmax: int, max degree
        :param option:
        :return: 3d-ndarray, indexes stand for (co-lat[rad], degree l, order m)
        """

        def a(nn, mm):
            return np.sqrt((2 * nn + 1) * (2 * nn - 1) / ((nn - mm) * (nn + mm)))

        def b(nn):
            return np.sqrt((2 * nn + 1) / (2 * nn))

        if option != 0:
            lat = (90. - lat) / 180. * np.pi

        pilm = MathTool.get_Legendre(lat, lmax, option=0)

        pilm_d = np.zeros_like(pilm)
        pilm_d[:, 0, 0] = 0
        pilm_d[:, 1, 0] = - np.sqrt(3) * np.sin(lat)
        pilm_d[:, 1, 1] = np.sqrt(3) * np.cos(lat)

        for n in range(2, lmax + 1):
            pilm_d[:, n, n] = b(n) * (
                    np.cos(lat) * pilm[:, n - 1, n - 1] + np.sin(lat) * pilm_d[:, n - 1, n - 1]
            )

            pilm_d[:, n, n - 1] = a(n, n - 1) * (
                    -np.sin(lat) * pilm[:, n - 1, n - 1] + np.cos(lat) * pilm_d[:, n - 1, n - 1]
            )

        for n in range(2, lmax + 1):
            for m in range(0, n - 1):
                pilm_d[:, n, m] = a(n, m) * (
                        -np.sin(lat) * pilm[:, n - 1, m] + np.cos(lat) * pilm_d[:, n - 1, m] - pilm_d[:, n - 2, m] / a(
                    n - 1, m)
                )

        return pilm_d

    @staticmethod
    def get_colat_lon_rad(lat, lon):
        """
        :param lat: geophysical coordinate in degree
        :param lon: geophysical coordinate in degree
        :return: co-latitude and longitude in rad
        """

        theta = (90. - lat) / 180. * np.pi
        phi = lon / 180. * np.pi

        return theta, phi

    @staticmethod
    def get_lat_lon_degree(theta, phi):
        """
        :param theta: co-latitude, geophysical coordinate in rad
        :param phi: longitude, geophysical coordinate in rad
        :return: latitude and longitude in degree
        """

        lat = 90. - theta * 180. / np.pi
        lon = phi * 180. / np.pi

        return lat, lon

    @staticmethod
    @DeprecationWarning
    def sort_covariance_matrix_old(cov_cs, lmax, lmin=2):
        """
        Generally the diagonal index of covariance matrix of spherical harmonic coefficients is sorted by degree like:
        ( var(c00); var(c10), var(c11), var(s11); var(c20), var(c21), var(s21), var(c22), var(s22); ... ),
        or omitted degree-0 and degree-1, like:
        ( var(c20), var(c21), var(s21), var(c22), var(s22); var(s33), var(s32), var(s31), var(c30), ... ),

        Due to the fact that the ranking of spherical harmonic coefficients in this program is based on:
        ( var(c00); var(s11), var(c10), var(c11); var(s22), var(s21), var(c20), var(c21), var(c22); ... ),
        it needs be re-sorted before get further used.

        :param cov_cs: 2d-array given in the above form
        :param lmax: max degree/order
        :param lmin: the minimum degree/order given in the input

        return: 2d-array that describe the re-sorted covariance matrix of spherical harmonic coefficients
        """
        if lmin == 0:
            cov_cs_full = cov_cs
        else:
            cov_cs_full = np.zeros(((lmax + 1) ** 2, (lmax + 1) ** 2))
            cov_cs_full[lmin ** 2:, lmin ** 2:] = cov_cs

        cov_cs_resorted = np.zeros_like(cov_cs_full)
        lrange = np.arange(1, 121 + 1, 2)

        def get_index2d_from_index1dnew(index):
            l = int(np.sqrt(index))
            m = index - l ** 2 - l
            return l, m

        def get_index1dold(l, m):
            return np.sum(lrange[:l]) - 1 + 2 * np.abs(m) + (1 if m <= 0 else 0)

        for index1 in range((lmax + 1) ** 2):
            l1, m1 = get_index2d_from_index1dnew(index1)
            index1_old = get_index1dold(l1, m1)

            for index2 in range((lmax + 1) ** 2):
                l2, m2 = get_index2d_from_index1dnew(index2)
                index2_old = get_index1dold(l2, m2)

                cov_cs_resorted[index1, index2] = cov_cs_full[index1_old, index2_old]

        return cov_cs_resorted

    @staticmethod
    def sort_covariance_matrix(cov_cs, lmax_input: int, lmin_input: int = 0):
        """
        Generally the diagonal index of covariance matrix of spherical harmonic coefficients is sorted by degree like:
        ( var(c00); var(c10), var(c11), var(s11); var(c20), var(c21), var(s21), var(c22), var(s22); ... ),
        or omitted degree-0 and degree-1, like:
        ( var(c20), var(c21), var(s21), var(c22), var(s22); var(s33), var(s32), var(s31), var(c30), ... ),

        Due to the fact that the ranking of spherical harmonic coefficients in this program is based on:
        ( var(c00); var(s11), var(c10), var(c11); var(s22), var(s21), var(c20), var(c21), var(c22); ... ),
        it needs be re-sorted before get further used.

        :param cov_cs: 2d-array given in the above form
        :param lmax_input: max degree/order
        :param lmin_input: the minimum degree/order given in the input

        return: 2d-array that describe the re-sorted covariance matrix of spherical harmonic coefficients
        """
        if lmin_input == 0:
            cov_cs_full = cov_cs
        else:
            cov_cs_full = np.zeros(((lmax_input + 1) ** 2, (lmax_input + 1) ** 2))
            cov_cs_full[lmin_input ** 2:, lmin_input ** 2:] = cov_cs

        lrange = np.arange(1, 121 + 1, 2)

        def get_index2d_from_index1dnew(index):
            l = int(np.sqrt(index))
            m = index - l ** 2 - l
            return l, m

        def get_index1dold(l, m):
            return np.sum(lrange[:l]) - 1 + 2 * np.abs(m) + (1 if m <= 0 else 0)

        identity_matrix = np.eye(np.shape(cov_cs_full)[0])
        elementary_transformation_matrix = np.zeros_like(cov_cs_full)

        '''index 1d (after transform) -> l, m -> index 1d (before transform)'''
        for index in range((lmax_input + 1) ** 2):
            l, m = get_index2d_from_index1dnew(index)
            index_old = get_index1dold(l, m)
            elementary_transformation_matrix[index] = identity_matrix[index_old]

        return elementary_transformation_matrix @ cov_cs_full @ elementary_transformation_matrix.T

    @staticmethod
    def get_design_matrix(function, t):
        m = len(t)
        arg_count = function.__code__.co_argcount - 1
        identity_m = np.eye(m)
        identity_m_extension = [list(identity_m[i]) * arg_count for i in range(m)]
        identity_m_extension = np.array(identity_m_extension).reshape(m * arg_count, m)

        t_extension = np.dot(identity_m_extension, t)
        identity_arg = np.eye(arg_count)
        arg_extension = [np.array(list(identity_arg[i]) * m) for i in range(arg_count)]

        A = function(t_extension, *arg_extension).reshape(m, arg_count)
        return A

    @staticmethod
    def dot_for_more(*mat):
        if len(mat) == 2:
            return np.dot(mat[0], mat[1])
        else:
            mat0 = mat[0]
            mat1 = mat[1]
            return MathTool.dot_for_more(np.dot(mat0, mat1), mat[2:])

    @staticmethod
    def curve_fit(function, t, *y, weight=None):
        """
        if weight is None, OLS;
        else, WLS.
        """

        if len(y) == 1:
            y = y[0].reshape(-1, 1)
        else:
            y = np.vstack(y)

        A = MathTool.get_design_matrix(function, t)

        if weight is not None:
            assert len(t) == len(weight)
            D = np.diag(np.sqrt(weight))
            A = D @ A
            y = D @ y

        AT = A.T
        # ATA_I = np.linalg.inv(np.dot(AT, A))
        ATA_I = np.linalg.pinv(np.dot(AT, A))
        A_ginv = np.dot(ATA_I, AT)
        A_ginv_A_ginv_T = np.dot(A_ginv, A_ginv.T)

        results = np.dot(A_ginv, y)

        epsilon = np.dot(A, results) - y

        if np.shape(epsilon)[1] == 1:
            var = np.var(epsilon) * A_ginv_A_ginv_T
        else:
            var = np.array([np.var(epsilon[:, i]) * A_ginv_A_ginv_T for i in range(np.shape(epsilon)[1])])

        return results.T, var

    @staticmethod
    def global_integral(grids, lat=None, lon=None, for_square=False):
        is_single = False
        if len(np.shape(grids)) == 2:
            grids = np.array([grids])

            is_single = True

        radius_e = GeoConstants.radius_earth

        grid_shape = np.shape(grids[0])

        if lat is None:
            lat = np.linspace(-90, 90, grid_shape[0])

        if lon is None:
            lon = np.linspace(-180, 180, grid_shape[1])

        colat_rad, lon_rad = MathTool.get_colat_lon_rad(lat, lon)

        dlat = np.abs(colat_rad[1] - colat_rad[0])
        dlon = np.abs(lon_rad[1] - lon_rad[0])

        domega = np.sin(colat_rad) * dlat * dlon * radius_e ** 2

        if for_square:
            integral = np.einsum('pij,i->p', grids, domega ** 2)
        else:
            integral = np.einsum('pij,i->p', grids, domega)

        if is_single:
            return integral[0]

        else:
            return integral

    @staticmethod
    def get_acreage(basin):

        assert len(basin.shape) in (2, 3)

        if len(basin.shape) == 2:
            acreage = MathTool.global_integral(np.array([basin]))[0]
        elif len(basin.shape) == 3:
            acreage = MathTool.global_integral(basin)
        else:
            assert False

        return acreage

    @staticmethod
    def xyz2grd(xyz):
        """
        xyz to grid
        :param xyz: np.ndarray or list, [ [lon0, lat0, z0], [lon1, lat1, z1], ... ], lons and lats' spacing are equal.
        :return: np.ndarray, grid
        """
        gs = max(np.abs(xyz[0][0] - xyz[1][0]), np.abs(xyz[0][1] - xyz[1][1]))
        lat, lon = np.zeros(int(180 / gs, )), np.zeros((int(360 / gs),))

        grid = np.zeros((int(180 / gs), int(360 / gs)))
        for i in xyz:
            this_lon, this_lat, this_value = i[0], i[1], i[2]

            l, m = MathTool.getGridIndex(this_lat, this_lon, gs)

            lat[l] = this_lat + gs / 2
            lon[m] = this_lon + gs / 2

            grid[l][m] = this_value
        return grid, lat, lon

    @staticmethod
    def getGridIndex(lat, lon, gs):
        return int((lat + 90) / gs), int((lon + 180) / gs)

    @staticmethod
    def getIndex(n: int, m: int):
        """
        index of Cnm at degree-ordered one-dimension array
        :param n: degree
        :param m: order
        :return:
        """
        assert m <= n

        return int(n * (n + 1) / 2 + m)

    @staticmethod
    def shrink(data, rows, cols):
        return data.reshape(rows, int(data.shape[0] / rows), cols, int(data.shape[1] / cols)).sum(axis=1).sum(axis=2)


if __name__ == '__main__':
    q = MathTool.get_cs_1d_index("s13,13")
    print(q)



from pysrc.auxiliary.load_file.LoadL2SH import LoadL2SH
from pysrc.auxiliary.preference.EnumClasses import L2ProductType, L2InstituteType


class GAXCorrectionConfig:
    def __init__(self):
        self.__GAX_type = L2ProductType.GAD
        self.__institute_type = L2InstituteType.CSR
        self.__dates = None

    def set_GAX_type(self, gax: L2ProductType):
        self.__GAX_type = gax
        return self

    def get_GAX_type(self):
        return self.__GAX_type

    def set_institute_type(self, institute: L2InstituteType):
        self.__institute_type = institute
        return self

    def get_institute_type(self):
        return self.__institute_type

    def set_dates(self, dates):
        self.__dates = dates
        return self

    def get_dates(self):
        return self.__dates

    pass


class GAXCorrection:
    def __init__(self):
        self.configuration = GAXCorrectionConfig()

    def apply_to(self, shc, shc_gax=None):
        if shc_gax is not None:
            pass

        else:
            lmax = shc.get_lmax()

            shc_gax = self.__load_gax(lmax)

        shc.value += shc_gax.cs

        return shc

    def __load_gax(self, lmax):
        dates = self.configuration.get_dates()
        begin_date = dates[0]
        end_date = dates[-1]

        institute = self.configuration.get_institute_type()
        product = self.configuration.get_GAX_type()

        load = LoadL2SH()

        load.configuration.set_begin_date(begin_date)
        load.configuration.set_end_date(end_date)
        load.configuration.set_institute(institute)
        load.configuration.set_product_type(product)
        load.configuration.set_lmax(lmax)

        shc_gax = load.get_shc()

        return shc_gax



"""
@Company: CGE-HUST, Wuhan, China
@Version: 1.0
@Author: Yang Fan
@Contact : yfan_cge@hust.edu.cn
@Modify Time: 2020/6/9 14:14
@Description:
"""
from multiprocessing import Pool

import numpy as np

from sagea.pysrc.post_processing.geometric_correction.old.GeoMathKit import GeoMathKit
from sagea.pysrc.post_processing.geometric_correction.old.LoveNumber import LoveNumber
from sagea.pysrc.post_processing.geometric_correction.old.RefEllipsoid import RefEllipsoid
from sagea.pysrc.post_processing.geometric_correction.old.Setting import SynthesisType, EllipsoidType, \
    LoveNumberType, Constants, HarAnalysisType


class Harmonic:
    """
    Harmonic analysis and synthesis: Ordinary 2D integration for computing Spherical Harmonic coefficients
    """

    def __init__(self, Ln: LoveNumber, Parallel: int = -1):
        self._Ln = Ln
        self._parallel = Parallel
        self._LoveMethod = LoveNumberType.AOD04
        self._ellipsoid = RefEllipsoid(EllipsoidType.gif48)
        pass

    def setLoveNumMethod(self, method: LoveNumberType):
        self._LoveMethod = method
        return self

    def setEllipsoid(self, ell: EllipsoidType):
        self._ellipsoid = RefEllipsoid(ell)
        return self

    @DeprecationWarning
    def analysis_old(self, Nmax: int, Inner: list, lat, lon, Pnm, kind=HarAnalysisType.InnerIntegral):
        """
        A simpson-type quadrature method.
        lat,lon, I and Pnm should keep accordance.
        The output is SH coefficients in terms of geo-potential.
        Notice: the ordering of Inner has to be N * M.  (rows sorted by latitude)

        Notice: the grid resolution has indeed determined the max degree (Nmax) of SH expansion, therefore the user
        defined Nmax can not be set without limitation, otherwise the high degree/order of output SH is fake.
        For regular grid that spans for N * 2N, the Nmax is suggested to be less than N/2-1.
        For Driscoll-healy grid that spans for N * N, the Nmax is suggested to be less than N/2-1.0
        For Gauss-Legendre quadrature grid that spans for (N+1)*(2*N+1), the Nmax is suggested to be less than N.

        :param kind: to let the function knows what type of the input gridded fields is
        :param Nmax: maximal degree of spherical harmonic expansion
        :param Inner: degree-dependent radial integration, len=Nmax or a simple gridded field len=1
                        dimension: [len, N*M]
        :param lat: latitude in degree   [dimension: N]
        :param lon: longitude in degree [dimension: M]
        :param Pnm: Legendre function
        :return: A set of spherical harmonic coefficients up to degree Nmax with one-dimension ordering type
        """
        # -------------------------------------------------------------
        # # extract the one-line lat and lon .
        # lat, lon = GeoMathKit.getCoLatLoninRad(lat, lon)
        #
        # # nlon = 2 * nlat
        # phi = lon[0:nlon]
        #
        # theta = np.zeros(nlat)
        # for i in range(nlat):
        #     theta[i] = lat[i * nlon]
        nlat = len(lat)
        nlon = len(lon)
        '''Co-latitude, longitude in [rad] '''
        theta, phi = GeoMathKit.getCoLatLoninRad(lat, lon)
        # --------------------------------------------------------------

        term1 = np.zeros(Nmax + 1)

        for l in range(0, Nmax + 1):
            term1[l] = 1 + 2 * l

        # kl = self._Ln.getNumber(Nmax, self._LoveMethod)
        # M = self._ellipsoid.Mass
        # R = self._ellipsoid.SemimajorAxis
        # factorSH = 4 * np.pi * R ** 2 / M * (1 + kl) / term1

        factorSH = self._factorHarAnalysis(Nmax=Nmax, kind=kind)

        # ---------------------------------------------------------------

        NMmax = int((Nmax + 1) * (Nmax + 2) / 2)

        factor1 = 2 * np.pi / nlon
        factor2 = 0.25 / nlat
        Cnm, Snm = np.zeros(NMmax), np.zeros(NMmax)

        if len(Inner) == 1:
            Am, Bm = np.zeros((Nmax + 1, nlat)), np.zeros((Nmax + 1, nlat))
            I_new = Inner[0].reshape(-1, nlon)

            for m in range(Nmax + 1):
                for i in range(nlat):
                    Am[m, i] = factor1 * np.sum(np.cos(m * phi) * I_new[i])
                    Bm[m, i] = factor1 * np.sum(np.sin(m * phi) * I_new[i])

            for n in range(Nmax + 1):
                for m in range(n + 1):
                    Cnm[GeoMathKit.getIndex(n, m)] = factorSH[n] * factor2 * np.sum(
                        Pnm[GeoMathKit.getIndex(n, m)] * np.sin(theta) * Am[m])
                    Snm[GeoMathKit.getIndex(n, m)] = factorSH[n] * factor2 * np.sum(
                        Pnm[GeoMathKit.getIndex(n, m)] * np.sin(theta) * Bm[m])

            return Cnm, Snm

        # -------------------------------------------------------------
        if len(Inner) > 1:
            assert len(Inner) >= (Nmax + 1)

        self.__input = (Nmax, nlat, nlon, Inner, factor1, phi, factorSH, Cnm, Snm, factor2, Pnm, theta)

        if self._parallel > 1:
            # for multiprocessing
            pool = Pool(self._parallel)
            ret = pool.map(self.job, list(range(Nmax + 1)))
            pool.close()
            pool.join()

            Cnm = []
            Snm = []

            ret.sort(key=lambda x: x[0])
            for i in range(Nmax + 1):
                Cnm = Cnm + ret[i][1]
                Snm = Snm + ret[i][2]

            return Cnm, Snm

        # for single process
        for n in range(Nmax + 1):
            self.job(n)

        return Cnm, Snm

    def analysis(self, Nmax: int, Inner: list, lat, lon, Pnm, kind=HarAnalysisType.InnerIntegral):
        """
        A simpson-type quadrature method.
        lat,lon, I and Pnm should keep accordance.
        The output is SH coefficients in terms of geo-potential.
        Notice: the ordering of Inner has to be N * M.  (rows sorted by latitude)

        Notice: the grid resolution has indeed determined the max degree (Nmax) of SH expansion, therefore the user
        defined Nmax can not be set without limitation, otherwise the high degree/order of output SH is fake.
        For regular grid that spans for N * N, the Nmax is suggested to be less than N/2.
        For Driscoll-healy grid that spans for N * 2N, the Nmax is suggested to be less than N/2.
        For Gauss-Legendre quadrature grid that spans for (N+1)*(2*N+1), the Nmax is suggested to be less than N.

        :param kind: to let the function knows what type of the input gridded fields is
        :param Nmax: maximal degree of spherical harmonic expansion
        :param Inner: degree-dependent radial integration, len=Nmax or a simple gridded field len=1
                        dimension: [len, N*M]
        :param lat: latitude in degree   [dimension: N]
        :param lon: longitude in degree [dimension: M]
        :param Pnm: Legendre function
        :return: A set of spherical harmonic coefficients up to degree Nmax with one-dimension ordering type
        """
        assert kind in HarAnalysisType
        # -------------------------------------------------------------
        # # extract the one-line lat and lon .
        # lat, lon = GeoMathKit.getCoLatLoninRad(lat, lon)
        #
        # # nlon = 2 * nlat
        # phi = lon[0:nlon]
        #
        # theta = np.zeros(nlat)
        # for i in range(nlat):
        #     theta[i] = lat[i * nlon]
        nlat = len(lat)
        nlon = len(lon)
        '''Co-latitude, longitude in [rad] '''
        theta, phi = GeoMathKit.getCoLatLoninRad(lat, lon)
        # --------------------------------------------------------------

        term1 = np.zeros(Nmax + 1)

        for l in range(0, Nmax + 1):
            term1[l] = 1 + 2 * l

        # kl = self._Ln.getNumber(Nmax, self._LoveMethod)
        # M = self._ellipsoid.Mass
        # R = self._ellipsoid.SemimajorAxis
        # factorSH = 4 * np.pi * R ** 2 / M * (1 + kl) / term1

        factorSH = self._factorHarAnalysis(Nmax=Nmax, kind=kind)

        # ---------------------------------------------------------------

        NMmax = int((Nmax + 1) * (Nmax + 2) / 2)

        factor1 = 2 * np.pi / nlon
        factor2 = 0.25 / nlat
        Cnm, Snm = np.zeros(NMmax), np.zeros(NMmax)

        if len(Inner) == 1:
            Am, Bm = np.zeros((Nmax + 1, nlat)), np.zeros((Nmax + 1, nlat))
            I_new = Inner[0].reshape(-1, nlon)

            for m in range(Nmax + 1):
                Am[m] = factor1 * np.array(I_new * np.asmatrix(np.cos(m * phi)).T).flatten()
                Bm[m] = factor1 * np.array(I_new * np.asmatrix(np.sin(m * phi)).T).flatten()

            thetaS = np.tile(np.sin(theta), (GeoMathKit.getIndex(Nmax, Nmax) + 1, 1))
            Qnm = Pnm * thetaS

            for n in range(Nmax + 1):
                indexM = np.arange(n + 1)
                Cnm[GeoMathKit.getIndex(n, 0):GeoMathKit.getIndex(n, n) + 1] = factorSH[n] * factor2 * \
                                                                               np.sum(Qnm[GeoMathKit.getIndex(n,
                                                                                                              0):GeoMathKit.getIndex(
                                                                                   n, n) + 1] * Am[indexM], 1)
                Snm[GeoMathKit.getIndex(n, 0):GeoMathKit.getIndex(n, n) + 1] = factorSH[n] * factor2 * \
                                                                               np.sum(Qnm[GeoMathKit.getIndex(n,
                                                                                                              0):GeoMathKit.getIndex(
                                                                                   n, n) + 1] * Bm[indexM], 1)
                pass

            return Cnm, Snm

        # -------------------------------------------------------------
        if len(Inner) > 1:
            assert len(Inner) >= (Nmax + 1)

        self.__input = (Nmax, nlat, nlon, Inner, factor1, phi, factorSH, Cnm, Snm, factor2, Pnm, theta)

        if self._parallel > 1:
            # for multiprocessing
            pool = Pool(self._parallel)
            ret = pool.map(self.job, list(range(Nmax + 1)))
            pool.close()
            pool.join()

            Cnm = []
            Snm = []

            ret.sort(key=lambda x: x[0])
            for i in range(Nmax + 1):
                Cnm = Cnm + ret[i][1]
                Snm = Snm + ret[i][2]

            return Cnm, Snm

        # for single process
        for n in range(Nmax + 1):
            self.job(n)

        return Cnm, Snm

    def analysis_new(self, Nmax: int, Gqij, lat, lon, PnmMat, kind=HarAnalysisType.InnerIntegral):
        """
        A simpson-type quadrature method.
        lat,lon, I and Pnm should keep accordance.
        The output is SH coefficients in terms of geo-potential.
        Notice: the ordering of Inner has to be N * M.  (rows sorted by latitude)

        Notice: the grid resolution has indeed determined the max degree (Nmax) of SH expansion, therefore the user
        defined Nmax can not be set without limitation, otherwise the high degree/order of output SH is fake.
        For regular grid that spans for N * N, the Nmax is suggested to be less than N/2.
        For Driscoll-healy grid that spans for N * 2N, the Nmax is suggested to be less than N/2.
        For Gauss-Legendre quadrature grid that spans for (N+1)*(2*N+1), the Nmax is suggested to be less than N.

        :param kind: to let the function knows what type of the input gridded fields is
        :param Nmax: maximal degree of spherical harmonic expansion
        :param Inner: degree-dependent radial integration, len=Nmax or a simple gridded field len=1
                        dimension: [len, N*M]
        :param lat: latitude in degree   [dimension: N]
        :param lon: longitude in degree [dimension: M]
        :param PnmMat: Legendre function, got from GeoMathKit.getPnmMatrix
        :return: A set of spherical harmonic coefficients up to degree Nmax with one-dimension ordering type
        """
        Gqij = np.array(Gqij)
        assert len(np.shape(Gqij)) in [3, 4]

        if len(np.shape(Gqij)) == 3:
            colat_rad, lon_rad = GeoMathKit.getCoLatLoninRad(lat, lon)
            nlat, nlon = len(colat_rad), len(lon_rad)

            factorSH = self._factorHarAnalysis(Nmax=Nmax, kind=kind)

            factor1 = np.ones((nlat, Nmax + 1))
            factor1[:, 0] += 1
            factor1 = 1 / (factor1 * nlon)

            factor2 = np.ones((Nmax + 1, Nmax + 1))
            factor2[:, 0] += 1
            factor2 *= np.pi / (2 * nlat)

            m = np.arange(Nmax + 1)
            g = (m[:, None] @ lon_rad[None, :]).T

            co = np.cos(g)
            so = np.sin(g)

            Am = np.einsum('pij,jm->pim', Gqij, co, optimize='greedy') * factor1
            Bm = np.einsum('pij,jm->pim', Gqij, so, optimize='greedy') * factor1

            Cnms = np.einsum('pim,ilm,i->plm', Am, PnmMat, np.sin(colat_rad), optimize='greedy') * factor2
            Snms = np.einsum('pim,ilm,i->plm', Bm, PnmMat, np.sin(colat_rad), optimize='greedy') * factor2

            Cnms = np.einsum('qlm,l->qlm', Cnms, factorSH, optimize='greedy')
            Snms = np.einsum('qlm,l->qlm', Snms, factorSH, optimize='greedy')

        else:
            Cnms = np.zeros((np.shape(Gqij)[1], Nmax + 1, Nmax + 1))
            Snms = np.zeros((np.shape(Gqij)[1], Nmax + 1, Nmax + 1))

            for l in range(Nmax + 1):
                this_Gqij = Gqij[l]
                Cqlm, Sqlm = self.analysis_new(Nmax, this_Gqij, lat, lon, PnmMat, kind=kind)
                Cnms[:, l, :l + 1] = Cqlm[:, l, :l + 1]
                Snms[:, l, :l + 1] = Sqlm[:, l, :l + 1]

        return Cnms, Snms

    def job(self, n):
        """
        Do not call this function as it is only used for parallel computing.
        :param n:
        :return:
        """
        Nmax, nlat, nlon, I, factor1, phi, factorSH, Cnm, Snm, factor2, Pnm, theta = self.__input

        Am, Bm = np.zeros((Nmax + 1, nlat)), np.zeros((Nmax + 1, nlat))
        I_new = I[n].reshape(-1, nlon)

        c = []
        s = []

        thetaS = np.tile(np.sin(theta), (GeoMathKit.getIndex(Nmax, Nmax) + 1, 1))
        Qnm = Pnm * thetaS

        for m in range(n + 1):
            Am[m] = factor1 * np.array(I_new * np.asmatrix(np.cos(m * phi)).T).flatten()
            Bm[m] = factor1 * np.array(I_new * np.asmatrix(np.sin(m * phi)).T).flatten()

        indexM = np.arange(n + 1)
        Cnm[GeoMathKit.getIndex(n, 0):GeoMathKit.getIndex(n, n) + 1] = factorSH[n] * factor2 * \
                                                                       np.sum(Qnm[GeoMathKit.getIndex(n,
                                                                                                      0):GeoMathKit.getIndex(
                                                                           n, n) + 1] * Am[indexM], 1)
        Snm[GeoMathKit.getIndex(n, 0):GeoMathKit.getIndex(n, n) + 1] = factorSH[n] * factor2 * \
                                                                       np.sum(Qnm[GeoMathKit.getIndex(n,
                                                                                                      0):GeoMathKit.getIndex(
                                                                           n, n) + 1] * Bm[indexM], 1)

        c = list(Cnm[GeoMathKit.getIndex(n, 0):GeoMathKit.getIndex(n, n) + 1])
        s = list(Snm[GeoMathKit.getIndex(n, 0):GeoMathKit.getIndex(n, n) + 1])
        return n, c, s

    @DeprecationWarning
    def job_old(self, n):
        """
        Do not call this function as it is only used for parallel computing.
        :param n:
        :return:
        """
        Nmax, nlat, nlon, I, factor1, phi, factorSH, Cnm, Snm, factor2, Pnm, theta = self.__input

        Am, Bm = np.zeros((Nmax + 1, nlat)), np.zeros((Nmax + 1, nlat))
        I_new = I[n].reshape(-1, nlon)

        c = []
        s = []
        for m in range(n + 1):
            for i in range(nlat):
                Am[m, i] = factor1 * np.sum(np.cos(m * phi) * I_new[i])
                Bm[m, i] = factor1 * np.sum(np.sin(m * phi) * I_new[i])

            Cnm[GeoMathKit.getIndex(n, m)] = factorSH[n] * factor2 * \
                                             np.sum(Pnm[GeoMathKit.getIndex(n, m)] * np.sin(theta) * Am[m])
            Snm[GeoMathKit.getIndex(n, m)] = factorSH[n] * factor2 * \
                                             np.sum(Pnm[GeoMathKit.getIndex(n, m)] * np.sin(theta) * Bm[m])

            c.append(Cnm[GeoMathKit.getIndex(n, m)])
            s.append(Snm[GeoMathKit.getIndex(n, m)])
        return n, c, s

    @DeprecationWarning
    def synthesis_old(self, Cnm, Snm, Nmax, lat, lon, kind=SynthesisType.Pressure):
        """
        A two step synthesis method, see the paper GJI (Nico Sneew)
        :param Cnm: in general, it should be the geo-potential coefficients sorted in one dimension.
        :param Snm:
        :param Nmax: Max degree of harmonic expansion.
        :param lat: geophysical latitude in unit "degree" [dimension : N]
        :param lon: geophysical latitude in unit "degree"[dimension : M]
        :param kind: define the form of synthesis
        :return: grid (nlat*nlon) [dimension N*M]
        """

        # -------------convert the (lat, lon) into mathematical form-----------
        lat, lon = GeoMathKit.getCoLatLoninRad(lat, lon)
        nlat = np.size(lat)
        nlon = np.size(lon)
        # -----------------get Pnm----------------------------------------------
        Pnm = GeoMathKit.getPnm(lat, Nmax)
        # ---------------Compute the Am and Bm---------------------------------
        factor = self.__factorSynthesis(Nmax, kind)
        Am, Bm = np.zeros((Nmax + 1, nlat)), np.zeros((Nmax + 1, nlat))
        for m in range(Nmax + 1):
            for l in range(m, Nmax + 1):
                index = GeoMathKit.getIndex(l, m)
                Am[m] = Am[m] + Pnm[index] * Cnm[index] * factor[l]
                Bm[m] = Bm[m] + Pnm[index] * Snm[index] * factor[l]

        Fout = np.zeros((nlon, nlat))

        for j in range(nlon):
            for m in range(Nmax + 1):
                Fout[j] = Fout[j] + Am[m] * np.cos(m * lon[j]) + Bm[m] * np.sin(m * lon[j])

        F = Fout.T

        return F

    def synthesis(self, Cnm, Snm, Nmax, lat, lon, kind=SynthesisType.Pressure):
        """
        A two-step synthesis method, see the paper GJI (Nico Sneew)
        :param Cnm: in general, it should be the geo-potential coefficients sorted in one dimension.
        :param Snm:
        :param Nmax: Max degree of harmonic expansion.
        :param lat: geophysical latitude in unit "degree" [dimension : N]
        :param lon: geophysical latitude in unit "degree"[dimension : M]
        :param kind: define the form of synthesis
        :return: grid (nlat*nlon) [dimension N*M]
        """
        assert kind in SynthesisType
        # -------------convert the (lat, lon) into mathematical form-----------
        lat, lon = GeoMathKit.getCoLatLoninRad(lat, lon)
        nlat = np.size(lat)
        nlon = np.size(lon)
        # -----------------get Pnm----------------------------------------------
        Pnm = GeoMathKit.getPnm(lat, Nmax)
        # ---------------Compute the Am and Bm---------------------------------
        factor = self.__factorSynthesis(Nmax, kind)
        Am, Bm = np.zeros((Nmax + 1, nlat)), np.zeros((Nmax + 1, nlat))
        for m in range(Nmax + 1):
            for l in range(m, Nmax + 1):
                index = GeoMathKit.getIndex(l, m)
                Am[m] = Am[m] + Pnm[index] * Cnm[index] * factor[l]
                Bm[m] = Bm[m] + Pnm[index] * Snm[index] * factor[l]

        Fout = 0

        for m in range(Nmax + 1):
            co = np.cos(m * lon)
            so = np.sin(m * lon)
            Fout = Fout + np.asmatrix(Am[m]).T * co + np.asmatrix(Bm[m]).T * so
            Fout = Fout + np.asmatrix(Am[m]).T * co + np.asmatrix(Bm[m]).T * so

        return np.array(Fout)

    def synthesis_new(self, Cqlm, Sqlm, Nmax, lat, lon, PnmMat, kind=SynthesisType.Pressure):
        """
        A two step synthesis method, see the paper GJI (Nico Sneew)
        :param Cnm: in general, it should be the geo-potential coefficients sorted in one dimension.
        :param Snm:
        :param Nmax: Max degree of harmonic expansion.
        :param lat: geophysical latitude in unit "degree" [dimension : N]
        :param lon: geophysical latitude in unit "degree"[dimension : M]
        :param kind: define the form of synthesis
        :return: grid (nlat*nlon) [dimension N*M]
        """
        colat_rad, lon_rad = GeoMathKit.getCoLatLoninRad(lat, lon)

        assert len(Cqlm) == len(Sqlm)
        Cnms = np.array(Cqlm)
        Snms = np.array(Sqlm)

        factor = self.__factorSynthesis(Nmax, kind)

        Am = np.einsum('ijk,ljk,j->ilk', Cnms, PnmMat, factor)
        Bm = np.einsum('ijk,ljk,j->ilk', Snms, PnmMat, factor)

        m = np.arange(Nmax + 1)
        g = m[:, None] @ lon_rad[None, :]

        co = np.cos(g)
        so = np.sin(g)

        Fout = Am @ co + Bm @ so

        return np.array(Fout)

    def __factorSynthesis(self, Nmax, kind=SynthesisType.synthesis):
        """
        Generate factor for SH coefficients while synthesising kinds of physical fields on the surface
        Notice: this function asks for that the SH coefficients has to be geo-potential ones.
        :param Nmax:
        :param kind:
        :return:
        """

        factor = np.zeros(Nmax + 1)
        termI = np.arange(Nmax + 1)
        term = 2 * termI + 1.

        R = self._ellipsoid.SemimajorAxis
        rho_ave = self._ellipsoid.rho_ave
        rp = R

        if kind == SynthesisType.synthesis:
            factor[:] = 1.
        elif kind == SynthesisType.Geoidheight:
            # factor[:] = R
            factor[:] = self._ellipsoid.GM / self._ellipsoid.SemimajorAxis / self._ellipsoid.je
        elif kind == SynthesisType.EWH:
            kl = self._Ln.getNumber(Nmax, self._LoveMethod)
            factor = (R * rho_ave / 3) * (term / (1 + kl)) / Constants.rho_water
        elif kind == SynthesisType.Pressure:
            kl = self._Ln.getNumber(Nmax, self._LoveMethod)
            factor = (R * rho_ave / 3) * (term / (1 + kl)) * Constants.g_wmo
        elif kind == SynthesisType.Potential:
            factor = self._ellipsoid.GM / rp * ((R / rp) ** termI)
        elif kind == SynthesisType.GravityDisturbance:
            factor = self._ellipsoid.GM / rp / rp * (termI + 1) * ((R / rp) ** termI)
        elif kind == SynthesisType.FreeAirGraviyAnomaly:
            factor = self._ellipsoid.GM / rp / rp * (termI - 1) * ((R / rp) ** termI)

        return factor

    def _factorHarAnalysis(self, Nmax, kind=HarAnalysisType.analysis):
        """
        Generate factor for SH coefficients while analyzing kinds of physical fields.
        Notice: this function asks for that the output SH coefficients should be geo-potential ones.
        :param Nmax:
        :param kind:
        :return:
        """

        factor = np.zeros(Nmax + 1)
        termI = np.arange(Nmax + 1)
        term = 2 * termI + 1.

        R = self._ellipsoid.SemimajorAxis
        rho_ave = self._ellipsoid.rho_ave
        M = self._ellipsoid.Mass

        if kind == HarAnalysisType.analysis:
            factor[:] = 1.
        elif kind == HarAnalysisType.GeoidHeight:
            # factor[:] = 1./R
            factor[:] = 1. / (self._ellipsoid.GM / self._ellipsoid.SemimajorAxis / self._ellipsoid.je)
        elif kind == HarAnalysisType.EWH:
            kl = self._Ln.getNumber(Nmax, self._LoveMethod)
            factor = 1. / ((R * rho_ave / 3) * (term / (1 + kl)) / Constants.rho_water)
            # factor = 4 * np.pi * R ** 2 / M * (1 + kl) / term * Constants.rho_water
        elif kind == HarAnalysisType.Pressure:
            kl = self._Ln.getNumber(Nmax, self._LoveMethod)
            factor = 1. / ((R * rho_ave / 3) * (term / (1 + kl)) * Constants.g_wmo)
        elif kind == HarAnalysisType.InnerIntegral:
            '''inner integral for only pressure type'''
            kl = self._Ln.getNumber(Nmax, self._LoveMethod)
            # factor = 1. / ((R * rho_ave / 3) * (term / (1 + kl)))
            '''equal to 4 * np.pi * R ** 2 / M * (1 + kl) / term1'''
            # factor = 4 * np.pi * R ** 2 / M * (1 + kl) / term
            factor = 1. / ((R * rho_ave / 3) * (term / (1 + kl)))
        elif kind == HarAnalysisType.InnerIntegral_EWH:
            '''inner integral for only pressure type'''
            kl = self._Ln.getNumber(Nmax, self._LoveMethod)
            # factor = 1. / ((R * rho_ave / 3) * (term / (1 + kl)))
            '''equal to 4 * np.pi * R ** 2 / M * (1 + kl) / term1'''
            # factor = 4 * np.pi * R ** 2 / M * (1 + kl) / term * Constants.rho_water
            factor = 1. / ((R * rho_ave / 3) * (term / (1 + kl)) / Constants.rho_water)

        return factor



"""
@Company: CGE-HUST, Wuhan, China
@Version: 1.0
@Author: Yang Fan
@Contact : yfan_cge@hust.edu.cn
@Modify Time: 2020/6/3 10:44
@Description:
"""

from enum import Enum


class FieldType(Enum):
    Pressure = 0
    EWH = 1


class Assumption(Enum):
    Sphere = 0
    Ellipsoid = 1
    ActualEarth = 2


class ForceFields(Enum):
    ERAinterim = 0
    ERA5 = 1


class IntegralChoice(Enum):
    GFZ06VI = 1
    GFZ05VI = 2
    GFZ04VI = 3
    EFVI = 4
    SphereSP = 5
    NonsphereSP = 6


class DataType(Enum):
    TEMP = 0
    SHUM = 1
    PSFC = 2
    PHISFC = 3


class InterpOption(Enum):
    Blockmean = 0
    Bilinear = 1
    IDW = 2
    Nearest = 3
    NoInterp = 4


class SynthesisType(Enum):
    """pure analysis or the geo-potential"""
    synthesis = 0
    '''Equivalent water height'''
    EWH = 1
    '''Geoid height (undulation)'''
    Geoidheight = 2
    '''Surface pressure'''
    Pressure = 3
    '''!Gravity disturbance, NOT the free air anomaly!!!'''
    GravityDisturbance = 4
    '''!Free air gravity anomaly, NOT the gravity disturbance!!!'''
    FreeAirGraviyAnomaly = 5
    '''!Potential (anomaly) - Blakely (7.2)'''
    Potential = 6


class HarAnalysisType(Enum):
    """pure analysis or the geo-potential"""
    analysis = 0
    '''Equivalent water height'''
    EWH = 1
    '''Geoid height'''
    GeoidHeight = 2
    '''Surface pressure'''
    Pressure = 3
    '''Inner integral: pressure/g'''
    InnerIntegral = 4
    '''Inner integral: ewh'''
    InnerIntegral_EWH = 5


class Constants:
    """gas constant for dry air"""
    Rd = 287.00
    # Rd = 287.06
    '''gravity constant g defined by WMO'''
    g_wmo = 9.80665
    # g_wmo = 9.7
    ''' water density'''
    rho_water = 1025.0


class EllipsoidType(Enum):
    """
    we provide two types of reference ellipsoid.
    """
    GRS80 = 0
    WGS84 = 1
    GRS80_IERS2010 = 2
    gif48 = 3  # mostly used for gravity field application


class LoveNumberType(Enum):
    PREM = 1
    AOD04 = 2
    Wang = 3
    IERS = 4


class AODtype(Enum):
    ATM = 0
    OCN = 1
    GLO = 2
    OBA = 3


class TidesType(Enum):
    S1 = 0
    S2 = 1
    S3 = 2
    M2 = 3
    P1 = 4
    K1 = 5
    N2 = 6
    L2 = 7
    T2 = 8
    R2 = 9
    T3 = 10
    R3 = 11


class L2ProductType(Enum):
    GAA = 0
    GAB = 1
    GAC = 2
    GAD = 3
    GSM = 4


class L2instituteType(Enum):
    GFZ = 0
    CSR = 1
    JPL = 2


class SAT(Enum):
    GRACE = 0
    GRACE_FO = 1


class setting:
    """
    system setting
    """



#!/use/bin/env python
# coding=utf-8
# @Author  : Shuhao Liu
# @Time    : 2025/6/13 10:28 
# @File    : __init__.py.py
import pathlib

import sagea
from sagea.pysrc.data_class.__SHC__ import SHC
from sagea.pysrc.data_class.__GRD__ import GRD

from sagea.pysrc.post_processing.harmonic.Harmonic import Harmonic

from sagea.pysrc.load_file.LoadL2SH import load_SHC as load_SHC
from sagea.pysrc.load_file.LoadL2LowDeg import load_low_degs as load_SHLow
from sagea.pysrc.load_file.LoadNoah import load_GLDAS_TWS
from sagea.pysrc.load_file.LoadShp import LoadShp

from sagea.pysrc.auxiliary.MathTool import MathTool
from sagea.pysrc.auxiliary.TimeTool import TimeTool
from sagea.pysrc.auxiliary.FileTool import FileTool
import sagea.pysrc.auxiliary.Preference as Preference

from sagea.pysrc.data_collection.collect_auxiliary import collect_auxiliary as collect_auxiliary_data


def set_auxiliary_data_path(path):
    path = pathlib.Path(path)
    assert path.exists(), f"Path {path} does not exist."

    sagea.pysrc.auxiliary.Preference.Config.aux_data_dir = pathlib.Path(path)



import datetime
from enum import Enum

import numpy as np


class TimeTool:
    class DateFormat(Enum):
        ClassDate = 1
        YearDay = 2
        MJD = 3
        YMD = 4
        YearFraction = 5
        TimeDelta = 6
        YMD_dash = 7

    @staticmethod
    def convert_date_format(date, input_type: DateFormat = DateFormat.ClassDate,
                            output_type: DateFormat = DateFormat.YearFraction,
                            from_date: datetime.date = None):
        """
        get the date in format of datetime.date
        input/output_type restrict the type or format of param date and the return.
        DateFormat.ClassDate -> datetime.date.
        DateFormat.YearDay -> str in format of 'yyyyddd', y: year; d: the d-th day of this year. e.g., '2002031'.
        DateFormat.MJD -> int, Modified Julian Day.
        DateFormat.YMD -> str in format of 'yyyymmdd', y: year; m:month; d: day. e.g., '20020131'.
        DateFormat.YMD_dash -> str in format of 'yyyy-mm-dd', y: year; m:month; d: day. e.g., '2002-01-31'.
        DateFormat.TimeDelta -> int, days from from_date.
        :return:
        """

        def _convert_to_class_date(d, i_type: TimeTool.DateFormat):

            if i_type == TimeTool.DateFormat.ClassDate:
                result = d

            elif i_type == TimeTool.DateFormat.YearDay:
                d_str = str(d)
                assert len(d_str) == 7
                year_str = d_str[:4]
                days_str = d_str[4:]

                result = datetime.date(int(year_str), 1, 1) + datetime.timedelta(int(days_str) - 1)

            elif i_type == TimeTool.DateFormat.MJD:
                d0 = datetime.date(1858, 11, 17)
                result = d0 + datetime.timedelta(days=int(d))

            elif i_type == TimeTool.DateFormat.YMD:
                d_str = str(d)
                assert len(d_str) == 8
                year_str = d_str[:4]
                month_str = d_str[4:6]
                day_str = d_str[6:]

                result = datetime.date(int(year_str), int(month_str), int(day_str))

            elif i_type == TimeTool.DateFormat.YMD_dash:
                d_str = str(d)
                assert len(d_str) in (8, 9, 10)
                ymd = d_str.split("-")
                assert len(ymd) == 3

                year_str = ymd[0]
                month_str = ymd[1]
                day_str = ymd[2]

                result = datetime.date(int(year_str), int(month_str), int(day_str))

            elif i_type == TimeTool.DateFormat.YearFraction:
                year = int(d)
                if TimeTool.is_leap(year):
                    days_of_year = int((d - year) * 366)
                else:
                    days_of_year = int((d - year) * 365)

                year_day = year * 1000 + days_of_year

                return _convert_to_class_date(year_day, i_type=TimeTool.DateFormat.YearDay)

            elif i_type == TimeTool.DateFormat.TimeDelta:
                assert from_date is not None
                return from_date + datetime.timedelta(days=int(d))

            else:
                raise Exception

            return result

        def _convert_from_class_date_to(d: datetime.date, o_type: TimeTool.DateFormat):
            year = d.year
            month = d.month
            day = d.day

            if o_type == TimeTool.DateFormat.ClassDate:
                result = d

            elif o_type == TimeTool.DateFormat.YearDay:
                time_delta = d - datetime.date(year, 1, 1)
                dth_day = time_delta.days + 1
                dth_day_str = str(dth_day).rjust(3, '0')

                result = f'{str(year)}{dth_day_str}'

            elif o_type == TimeTool.DateFormat.MJD:
                timedelta = d - datetime.date(1858, 11, 17)
                result = timedelta.days

            elif o_type == TimeTool.DateFormat.YMD:
                year_str = str(year)
                month_str = str(month).rjust(2, '0')
                day_str = str(day).rjust(2, '0')
                result = f'{year_str}{month_str}{day_str}'

            elif o_type == TimeTool.DateFormat.YMD_dash:
                year_str = str(year)
                month_str = str(month).rjust(2, '0')
                day_str = str(day).rjust(2, '0')
                result = f'{year_str}-{month_str}-{day_str}'

            elif o_type == TimeTool.DateFormat.YearFraction:
                days_of_year = (d - datetime.date(year, 1, 1)).days + 0.5

                if TimeTool.is_leap(year):
                    return year + days_of_year / 366

                else:
                    return year + days_of_year / 365

            elif o_type == TimeTool.DateFormat.TimeDelta:
                assert from_date is not None
                return (d - from_date).days

            else:
                raise Exception

            return result

        # if type(date) is list:
        if len(np.shape(date)) == 1:
            d_class_date = [_convert_to_class_date(date[i], i_type=input_type) for i in range(len(date))]
            d_target_type = [_convert_from_class_date_to(d_class_date[i], o_type=output_type) for i in
                             range(len(d_class_date))]

        else:
            d_class_date = _convert_to_class_date(date, i_type=input_type)
            d_target_type = _convert_from_class_date_to(d_class_date, o_type=output_type)

        return d_target_type

    @staticmethod
    def get_the_final_day_of_this_month(date: datetime.date = None, year: int = None, month: int = None):
        input_date = date is not None

        input_year = year is not None
        input_month = month is not None

        assert input_year == input_month
        assert input_date ^ input_year

        if input_date:
            year = date.year
            month = date.month
        else:
            pass

        if month == 12:
            return datetime.date(year, month, 31)

        else:
            return datetime.date(year, month + 1, 1) - datetime.timedelta(1)

    @staticmethod
    def is_leap(year):
        if (year % 4 == 0) and (year % 100 != 0 or year % 400 == 0):
            return True
        else:
            return False

    @staticmethod
    def get_average_dates(begin, end, unused=None):
        """
        Obtain the average date by entering the start and end date(s).
        The start/end date(s) can be a datetime.date object, or a list composed of it with consistent length,
        and the return value can also be a datetime.date object or a list composed of it.

        The optional parameter 'unused' is supposed to be a list, or a list nested within a list,
        whose elements should also be a datetime.date object.
        If there are parameters passed in of parameter 'unused',
        these dates will not be considered when calculating the average.
        """

        assert type(begin) in (datetime.date, list)
        assert type(end) in (datetime.date, list)
        assert type(end) is type(begin)

        if type(begin) is datetime.date:
            if unused is None:
                time_delta = end - begin

                ave_date = begin + time_delta / 2
                return ave_date

            else:
                '''get used dates'''
                used_days = []
                max_days = (end - begin).days
                for i in range(max_days + 1):
                    this_date = begin + datetime.timedelta(days=i)
                    if this_date not in unused:
                        used_days.append(this_date)

                '''get averaged dates of used dates'''
                mjd = 0
                for i in range(len(used_days)):
                    mjd += TimeTool.convert_date_format(
                        used_days[i],
                        input_type=TimeTool.DateFormat.ClassDate,
                        output_type=TimeTool.DateFormat.MJD
                    )

                return TimeTool.convert_date_format(
                    mjd / len(used_days),
                    input_type=TimeTool.DateFormat.MJD,
                    output_type=TimeTool.DateFormat.ClassDate
                )

        else:
            ave_dates = []
            for i in range(len(begin)):
                ave_date = TimeTool.get_average_dates(begin[i], end[i], unused[i] if type(unused) is list else None)

                ave_dates.append(ave_date)

            return ave_dates



import numpy as np

from sagea.pysrc.auxiliary.Preference import Dimension
from sagea.pysrc.auxiliary.MathTool import MathTool


class Harmonic:
    """
    Harmonic analysis and synthesis: Ordinary 2D integration for computing Spherical Harmonic coefficients
    """

    def __init__(self, lat, lon, lmax: int, option=0):
        """

        :param lat: Co-latitude if option=0, unit[rad]; geophysical latitude if option = others, unit[degree]
        :param lon: If option=0, unit[rad]; else unit[degree]
        :param lmax: int, max degree/order
        :param option:
        """
        if option != 0:
            self.lat, self.lon = MathTool.get_colat_lon_rad(lat, lon)
        else:
            self.lat, self.lon = lat, lon

        self.lmax = lmax

        self._prepare()

    def _prepare(self):
        """pre-process some parameters of the 'two-step' method."""
        self.nlat, self.nlon = len(self.lat), len(self.lon)

        self.pilm = MathTool.get_Legendre(self.lat, self.lmax)
        # Associative Legendre polynomials, indexes stand for (co-lat[rad], degree l, order m)
        # 3-d array [theta, l, m] shape: (nlat * (lmax + 1) * (lmax + 1))

        # self.si = np.sin(self.lat) * np.pi / (2 * self.lmax + 1)

        # get Neumann weights
        # x_mat = np.array([np.cos(self.lat) ** i for i in range(self.nlat)])
        # r_vec = np.ones(self.nlat) * 2 / np.arange(1, self.nlat + 1, 1)
        # r_vec[np.arange(1, self.nlat + 1, 2)] = 0
        # self.wi = np.linalg.pinv(x_mat) @ r_vec

        m = np.arange(self.lmax + 1)
        self.g = m[:, None] @ self.lon[None, :]

        self.factor1 = np.ones((self.nlat, self.lmax + 1))
        self.factor1[:, 0] += 1
        self.factor1 = 1 / (self.factor1 * self.nlon)

        self.factor2 = np.ones((self.lmax + 1, self.lmax + 1))
        self.factor2[:, 0] += 1
        self.factor2 *= np.pi / (2 * self.nlat)

        self.factor3 = np.ones((self.lmax + 1, self.lmax + 1))
        self.factor3[:, 0] += 1
        self.factor3 /= 2
        pass

    def analysis(self, gqij: np.ndarray, special_type: Dimension = None):
        assert len(gqij.shape) in (2, 3)

        single = (len(gqij.shape) == 2)
        if single:
            gqij = np.array([gqij])

        assert special_type in (
            Dimension.HorizontalDisplacementEast, Dimension.HorizontalDisplacementNorth, None)

        if special_type in (
                Dimension.HorizontalDisplacementEast, Dimension.HorizontalDisplacementNorth):
            assert False, "Horizontal Displacement is not supported yet."

        g = self.g
        co = np.cos(g)  # cos(m phi)
        so = np.sin(g)  # sin(m phi)

        am = np.einsum('pij,mj->pim', gqij, co, optimize='greedy') * self.factor1
        bm = np.einsum('pij,mj->pim', gqij, so, optimize='greedy') * self.factor1

        if special_type is None:
            cqlm = np.einsum('pim,ilm,i->plm', am, self.pilm, np.sin(self.lat), optimize='greedy') * self.factor2
            sqlm = np.einsum('pim,ilm,i->plm', bm, self.pilm, np.sin(self.lat), optimize='greedy') * self.factor2

            # cqlm = np.einsum('pim,ilm,i->plm', am, self.pilm, self.wi, optimize='greedy') * self.factor3
            # sqlm = np.einsum('pim,ilm,i->plm', bm, self.pilm, self.wi, optimize='greedy') * self.factor3

            # cqlm[:, :, 0] = (np.einsum('pim,ilm->plm', am * self.wi[:, None], self.pilm,
            #                               optimize='greedy') * self.factor3)[:, :, 0]
            # sqlm[:, :, 0] = (np.einsum('pim,ilm->plm', bm * self.wi[:, None], self.pilm,
            #                            optimize='greedy') * self.factor3)[:, :, 0]

            # cqlm = np.einsum('pim,ilm->plm', am * self.wi[:, None], self.pilm,
            #                  optimize='greedy') * self.factor3
            # sqlm = np.einsum('pim,ilm->plm', bm * self.wi[:, None], self.pilm,
            #                  optimize='greedy') * self.factor3

        elif special_type == Dimension.HorizontalDisplacementNorth:
            """do NOT use this code!"""

            pilm_derivative = MathTool.get_Legendre_derivative(self.lat, self.lmax)

            cqlm = np.einsum('pim,ilm,i->plm', -am, pilm_derivative, np.sin(self.lat)) * self.factor2
            sqlm = np.einsum('pim,ilm,i->plm', -bm, pilm_derivative, np.sin(self.lat)) * self.factor2

        elif special_type == Dimension.HorizontalDisplacementEast:
            """do NOT use this code!"""

            mrange = np.arange(self.lmax + 1)
            pilm_divide_sin_theta = self.pilm / np.sin(self.lat)[:, None, None]
            m_pilm_divide_sin_theta = np.einsum("m,ilm->ilm", mrange, pilm_divide_sin_theta)

            am_east = np.einsum('pij,mj->pim', gqij, -so, optimize='greedy') * self.factor1
            bm_east = np.einsum('pij,mj->pim', gqij, co, optimize='greedy') * self.factor1

            cqlm = np.einsum('pim,ilm,i->plm', am_east, m_pilm_divide_sin_theta, np.sin(self.lat)) * self.factor2
            sqlm = np.einsum('pim,ilm,i->plm', bm_east, m_pilm_divide_sin_theta, np.sin(self.lat)) * self.factor2

        else:
            assert False

        if single:
            assert cqlm.shape[0] == 1 and sqlm.shape[0] == 1
            return cqlm[0], sqlm[0]

        else:
            return cqlm, sqlm

    def synthesis(self, cqlm: np.ndarray, sqlm: np.ndarray, special_type: Dimension = None):
        assert cqlm.shape == sqlm.shape
        assert len(cqlm.shape) in (2, 3)
        single = (len(cqlm.shape) == 2)
        if single:
            cqlm = np.array([cqlm])
            sqlm = np.array([sqlm])

        assert special_type in (
            Dimension.HorizontalDisplacementEast, Dimension.HorizontalDisplacementNorth, None)

        cqlm = np.array(cqlm)
        sqlm = np.array(sqlm)

        if special_type is None:
            am = np.einsum('ijk,ljk->ilk', cqlm, self.pilm)
            bm = np.einsum('ijk,ljk->ilk', sqlm, self.pilm)

        elif special_type is Dimension.HorizontalDisplacementNorth:
            pilm_derivative = MathTool.get_Legendre_derivative(self.lat, self.lmax)

            am = np.einsum('ijk,ljk->ilk', -cqlm, pilm_derivative)
            bm = np.einsum('ijk,ljk->ilk', -sqlm, pilm_derivative)

        elif special_type is Dimension.HorizontalDisplacementEast:
            pilm_divide_sin_theta = self.pilm / np.sin(self.lat)[:, None, None]
            mrange = np.arange(self.lmax + 1)

            am = np.einsum('ijk,k,ljk->ilk', sqlm, mrange, pilm_divide_sin_theta)
            bm = np.einsum('ijk,k,ljk->ilk', -cqlm, mrange, pilm_divide_sin_theta)

        else:
            assert False

        co = np.cos(self.g)
        so = np.sin(self.g)

        gqij = am @ co + bm @ so

        if single:
            assert gqij.shape[0] == 1
            return gqij[0]
        else:
            return gqij



import datetime

import numpy as np
from netCDF4 import Dataset

from sagea.pysrc.auxiliary import FileTool
from sagea.pysrc.auxiliary.MathTool import MathTool
from sagea.pysrc.data_class.__GRD__ import GRD


class LoadNOAH21:
    def __init__(self):
        self.nc = None
        self.keys = None

    def setFile(self, file):
        self.nc = Dataset(file)
        self.keys = self.nc.variables.keys()
        return self

    def get2dData(self, key, full_lat=False):
        assert key in self.keys, 'no such key'
        data = np.array(self.nc.variables[key])[0]
        xyz = []
        for i in range(len(data)):
            for j in range(len((data[i]))):
                lat = int(i - 90) + 30
                lon = j - 180
                xyz.append([lon, lat, data[i][j]])
        xyz = np.array(xyz)
        grid, lat, lon = MathTool.xyz2grd(xyz)

        if full_lat:
            lat = np.arange(-90 + 0.5, 90 + 0.5, 1)

        return grid, lat, lon


def load_GLDAS_TWS(filepath, full_lat=True, components: iter = None):
    """
    The GLDAS/Noah soil moisture (SM), snow water equivalent (SWE), and plant canopy water storage (PCSW) are jointly
    used to calculate the TWS variations. doi: 10.1155/2019/3874742
    :param filepath: path + filename.nc of NOAH
    :param full_lat:
    :param components: iter of strings, defaults to []
    :return: 1*1 degree TWS map [m]
    """
    if components is None:
        components = ["SoilMoi0_10cm_inst", "SoilMoi10_40cm_inst", "SoilMoi40_100cm_inst", "SoilMoi100_200cm_inst",
                      "CanopInt_inst", "SWE_inst"]

    nc = LoadNOAH21().setFile(filepath)
    total, lat, lon = None, None, None
    for i in range(len(components)):
        this_component = components[i]

        if total is None:
            total, lat, lon = nc.get2dData(this_component, full_lat=full_lat)
        else:
            total += nc.get2dData(this_component, full_lat=full_lat)[0]

    assert total is not None

    return total / 1000, lat, lon



"""
This file refers to ggtools package written by lcx366 2016.
For more information, please refer to https://github.com/lcx366/GGTOOLS
"""
import pathlib
import struct
import sys
import numpy as np
from sagea.pysrc.auxiliary import Preference
from scipy.linalg import block_diag

from enum import Enum

from sagea.pysrc.data_collection.collect_auxiliary import collect_ddk_data
from sagea.pysrc.post_processing.filter.Base import SHCFilter


class DDKFilterType(Enum):
    DDK1 = 1
    DDK2 = 2
    DDK3 = 3
    DDK4 = 4
    DDK5 = 5
    DDK6 = 6
    DDK7 = 7
    DDK8 = 8


def read_BIN(file, mode='packed'):
    if mode == 'packed':
        unpack = False
    elif mode == 'full':
        unpack = True
    else:
        raise Exception("Only 'packed' or 'full' are avaliable.")

    endian = sys.byteorder

    if endian == 'little':
        f = open(file, 'rb')
    else:
        raise Exception('The endian of the binary file is little, but the endian of OS is big.')

    dat = {}
    dat['version'] = f.read(8).decode().strip()
    dat['type'] = f.read(8).decode()
    dat['descr'] = f.read(80).decode().strip()

    for key in ['nints', 'ndbls', 'nval1', 'nval2']:
        dat[key] = struct.unpack('<I', f.read(4))[0]

    for key in ['pval1', 'pval2']:
        dat[key] = struct.unpack('<I', f.read(4))[0]

    dat['nvec'], dat['pval2'] = 0, 1
    dat['nread'], dat['nval2'] = 0, dat['nval1']

    nblocks = struct.unpack('<i', f.read(4))[0]

    lists = f.read(dat['nints'] * 24).decode().split()
    for element in lists:
        dat[element] = struct.unpack('<i', f.read(4))[0]

    lists = f.read(dat['ndbls'] * 24).decode().replace(':', '').split()
    for element in lists:
        dat[element] = struct.unpack('<d', f.read(8))[0]

    lists = f.read(dat['nval1'] * 24).decode()
    dat['side1_d'] = [(lists[i:i + 24]).replace('         ', '') for i in range(0, len(lists), 24)]

    dat['blockind'] = np.array(struct.unpack('<' + str(nblocks) + 'i', f.read(4 * nblocks)))

    dat['side2_d'] = dat['side1_d']

    npack1 = dat['pval1'] * dat['pval2']
    dat['pack1'] = np.array(struct.unpack('<' + str(npack1) + 'd', f.read(8 * npack1)))

    f.close()

    if not unpack: return dat

    sz = dat['blockind'][0]
    dat['mat1'] = dat['pack1'][:sz ** 2].reshape(sz, sz).T

    shift1 = shift2 = sz ** 2

    for i in range(1, nblocks):
        sz = dat['blockind'][i] - dat['blockind'][i - 1]
        shift2 = shift1 + sz ** 2
        dat['mat1'] = block_diag(dat['mat1'], dat['pack1'][shift1:shift2].reshape(sz, sz).T)
        shift1 = shift2
    del dat['pack1']

    return dat


def filterSH(W, cilm, cilm_std=None):
    lmax = cilm.shape[1] - 1

    lmaxfilt, lminfilt = W['Lmax'], W['Lmin']

    lmaxout = min(lmax, lmaxfilt)

    cilm_filter = np.zeros_like(cilm)
    cilm_std_filter = np.zeros_like(cilm_std)

    lastblckind, lastindex = 0, 0

    for iblk in range(W['Nblocks']):
        degree = (iblk + 1) // 2

        if degree > lmaxout: break
        trig = (iblk + int(iblk > 0) + 1) % 2

        sz = W['blockind'][iblk] - lastblckind

        blockn = np.identity(lmaxfilt + 1 - degree)

        lminblk = max(lminfilt, degree)

        shift = lminblk - degree
        blockn[shift:, shift:] = W['pack1'][lastindex:lastindex + sz ** 2].reshape(sz, sz).T

        if trig:
            cilm_filter[0, degree:lmaxout + 1, degree] = np.dot(blockn[:lmaxout + 1 - degree, :lmaxout + 1 - degree],
                                                                cilm[0, degree:lmaxout + 1, degree])
        else:
            cilm_filter[1, degree:lmaxout + 1, degree] = np.dot(blockn[:lmaxout + 1 - degree, :lmaxout + 1 - degree],
                                                                cilm[1, degree:lmaxout + 1, degree])

        if cilm_std is not None:
            if trig:
                cilm_std_filter[0, degree:lmaxout + 1, degree] = np.sqrt(
                    np.dot(blockn[:lmaxout + 1 - degree, :lmaxout + 1 - degree] ** 2,
                           cilm_std[0, degree:lmaxout + 1, degree] ** 2))
            else:
                cilm_std_filter[1, degree:lmaxout + 1, degree] = np.sqrt(
                    np.dot(blockn[:lmaxout + 1 - degree, :lmaxout + 1 - degree] ** 2,
                           cilm_std[1, degree:lmaxout + 1, degree] ** 2))

        lastblckind = W['blockind'][iblk]
        lastindex = lastindex + sz ** 2
        pass

    if cilm_std is None:
        return cilm_filter
    else:
        return cilm_filter, cilm_std_filter


class DDKConfig:
    def __init__(self):
        self.filter_type = DDKFilterType.DDK3

    def set_filter_type(self, filter_type: DDKFilterType):

        if type(filter_type) is DDKFilterType:
            self.filter_type = filter_type

        elif type(filter_type) is int:
            ddk_type = DDKFilterType.__members__[f'DDK{str(filter_type)}']
            self.filter_type = ddk_type

        else:
            return -1

        return self


class DDK(SHCFilter):
    def __init__(self):
        self.configuration = DDKConfig()

    def __apply_to_csqlm(self, cqlm, sqlm):
        ddktype = self.configuration.filter_type

        self.check_files()

        ddk_data_dir = Preference.Config.aux_data_dir / "ddk_data"

        if ddktype == DDKFilterType.DDK1:
            read_path = ddk_data_dir / 'Wbd_2-120.a_1d14p_4'
        elif ddktype == DDKFilterType.DDK2:
            read_path = ddk_data_dir / 'Wbd_2-120.a_1d13p_4'
        elif ddktype == DDKFilterType.DDK3:
            read_path = ddk_data_dir / 'Wbd_2-120.a_1d12p_4'
        elif ddktype == DDKFilterType.DDK4:
            read_path = ddk_data_dir / 'Wbd_2-120.a_5d11p_4'
        elif ddktype == DDKFilterType.DDK5:
            read_path = ddk_data_dir / 'Wbd_2-120.a_1d11p_4'
        elif ddktype == DDKFilterType.DDK6:
            read_path = ddk_data_dir / 'Wbd_2-120.a_5d10p_4'
        elif ddktype == DDKFilterType.DDK7:
            read_path = ddk_data_dir / 'Wbd_2-120.a_1d10p_4'
        elif ddktype == DDKFilterType.DDK8:
            read_path = ddk_data_dir / 'Wbd_2-120.a_5d9p_4'
        else:
            raise Exception

        Wbd = read_BIN(read_path)

        assert len(cqlm) == len(sqlm)
        cilms_filtered = []
        for i in range(len(cqlm)):
            cilm = np.array([cqlm[i], sqlm[i]])
            cilms_filtered.append(filterSH(Wbd, cilm))  # list, [ [c1,s1], [c2,s2], ... ]

        cqlm = np.array([cilms_filtered[i][0] for i in range(len(cilms_filtered))])
        sqlm = np.array([cilms_filtered[i][1] for i in range(len(cilms_filtered))])
        return cqlm, sqlm

    def apply_to(self, cqlm, sqlm):
        cqlm, sqlm, single = self._cs_to_3d_array(cqlm, sqlm)

        cqlm_f, sqlm_f = self.__apply_to_csqlm(cqlm, sqlm)

        if single:
            assert cqlm_f.shape[0] == sqlm_f.shape[0] == 1
            return cqlm_f[0], sqlm_f[0]
        else:
            return cqlm_f, sqlm_f

    def check_files(self):
        check_flag = True

        aux_dir = Preference.Config.aux_data_dir
        ddk_data_dir = pathlib.Path(aux_dir) / "ddk_data"

        for filename in (
                'Wbd_2-120.a_1d14p_4',  # DDK1
                'Wbd_2-120.a_1d13p_4',  # DDK2
                'Wbd_2-120.a_1d12p_4',  # DDK3
                'Wbd_2-120.a_5d11p_4',  # DDK4
                'Wbd_2-120.a_1d11p_4',  # DDK5
                'Wbd_2-120.a_5d10p_4',  # DDK6
                'Wbd_2-120.a_1d10p_4',  # DDK7
                'Wbd_2-120.a_5d9p_4',  # DDK8
        ):
            if not (ddk_data_dir / filename).exists():
                check_flag = False

                break

        if not check_flag:
            self.__download_files()

    @staticmethod
    def __download_files():
        try:
            collect_ddk_data()

        except Exception as e:
            print("download DDK data failed, please try again later, or manually download the auxiliary data "
                  "using 'sagea.collect_auxiliary()' and set auxiliary path "
                  "using 'sagea.set_auxiliary_data_path()'.")

            raise e
